from collections.abc import Mapping
from copy import copy
from pathlib import Path
from typing import Any, cast

import click
from click.core import ParameterSource
from dagster_dg_core.component import EnvRegistry
from dagster_dg_core.config import (
    get_config_from_cli_context,
    has_config_on_cli_context,
    normalize_cli_config,
    set_config_on_cli_context,
)
from dagster_dg_core.context import DgContext
from dagster_dg_core.shared_options import GLOBAL_OPTIONS, dg_global_options
from dagster_dg_core.utils import (
    DgClickCommand,
    DgClickGroup,
    exit_with_error,
    generate_missing_registry_object_error_message,
    json_schema_property_to_click_option,
    not_none,
    parse_json_option,
)
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper
from dagster_shared import check
from dagster_shared.error import DagsterUnresolvableSymbolError
from dagster_shared.serdes.objects import EnvRegistryKey, EnvRegistryObjectSnap
from dagster_shared.seven import load_module_object

from dagster_dg_cli.scaffold import (
    ScaffoldFormatOptions,
    scaffold_inline_component,
    scaffold_registry_object,
)

# These commands are not dynamically generated, but perhaps should be.
HARDCODED_COMMANDS = {"component"}


# The `dg scaffold defs` command is special because its subcommands are dynamically generated
# from the available components in the environment. Because the component types
# depend on the component modules we are using, we cannot resolve them until we have know these
# component modules, which can be set via the `--use-component-module` option, e.g.
#
#     dg --use-component-module dagster_components.test ...
#
# To handle this, we define a custom click.Group subclass that loads the commands on demand.
class ScaffoldDefsGroup(DgClickGroup):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands_defined = False

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        if not self._commands_defined and cmd_name not in HARDCODED_COMMANDS:
            self._define_commands(ctx)

        # First try exact match
        cmd = super().get_command(ctx, cmd_name)
        return cmd or self._get_matching_command(ctx, cmd_name)

    def list_commands(self, ctx: click.Context) -> list[str]:
        if not self._commands_defined:
            self._define_commands(ctx)
        return super().list_commands(ctx)

    def _define_commands(self, cli_context: click.Context) -> None:
        """Dynamically define a command for each registered component type."""
        if not has_config_on_cli_context(cli_context):
            cli_context.invoke(not_none(self.callback), **cli_context.params)
        config = get_config_from_cli_context(cli_context)
        dg_context = DgContext.from_file_discovery_and_command_line_config(Path.cwd(), config)

        registry = EnvRegistry.from_dg_context(dg_context)

        # Keys where the actual class name is not shared with any other key will use the class name
        # as a command alias.
        keys_by_name: dict[str, set[EnvRegistryKey]] = {}
        for key in registry.keys():
            keys_by_name.setdefault(key.name, set()).add(key)

        for key, component_type in registry.items():
            self._create_subcommand(
                key, component_type, use_typename_alias=len(keys_by_name[key.name]) == 1
            )

        self._commands_defined = True

    def _create_subcommand(
        self,
        key: EnvRegistryKey,
        obj: EnvRegistryObjectSnap,
        use_typename_alias: bool,
    ) -> None:
        # We need to "reset" the help option names to the default ones because we inherit the parent
        # value of context settings from the parent group, which has been customized.
        aliases = [
            *[alias.to_typename() for alias in obj.aliases],
            *([key.name] if use_typename_alias else []),
        ]

        @self.command(
            cls=ScaffoldDefsSubCommand,
            name=key.to_typename(),
            context_settings={"help_option_names": ["-h", "--help"]},
            aliases=aliases,
            help=obj.description or obj.summary,
        )
        @click.argument("defs_path", type=str)
        @click.pass_context
        @cli_telemetry_wrapper
        def scaffold_command(
            cli_context: click.Context,
            defs_path: str,
            **other_opts: Any,
        ) -> None:
            f"""Scaffold a {key.name} object.

            This command must be run inside a Dagster project directory. The component scaffold will be
            placed in submodule `<project_name>.defs.<INSTANCE_NAME>`.

            Objects can optionally be passed scaffold parameters. There are two ways to do this:

            (1) Passing a single --json-params option with a JSON string of parameters. For example:

                dg scaffold foo.bar my_object --json-params '{{"param1": "value", "param2": "value"}}'`.

            (2) Passing each parameter as an option. For example:

                dg scaffold foo.bar my_object --param1 value1 --param2 value2`

            It is an error to pass both --json-params and key-value pairs as options.
            """
            cli_config = get_config_from_cli_context(cli_context)
            dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)

            # json_params will not be present in the key_value_params if no scaffold properties
            # are defined.
            json_scaffolder_params = other_opts.pop("json_params", None)

            # format option is only present if we are dealing with a component. Otherewise we
            # default to python for decorator scaffolding. Default is YAML (set by option) for
            # components.
            scaffolder_format = cast("ScaffoldFormatOptions", other_opts.pop("format", "python"))

            # Remanining options are scaffolder params
            key_value_scaffolder_params = other_opts

            check.invariant(
                scaffolder_format in ["yaml", "python"],
                "format must be either 'yaml' or 'python'",
            )
            _core_scaffold(
                dg_context,
                cli_context,
                key,
                defs_path,
                key_value_scaffolder_params,
                scaffolder_format,
                json_scaffolder_params,
            )

        if obj.is_component:
            scaffold_command.params.append(
                click.Option(
                    ["--format"],
                    type=click.Choice(["yaml", "python"], case_sensitive=False),
                    default="yaml",
                    help="Format of the component configuration (yaml or python)",
                )
            )

        # If there are defined scaffold properties, add them to the command. Also only add
        # `--json-params` if there are defined scaffold properties.
        if obj.scaffolder_schema and obj.scaffolder_schema.get("properties"):
            scaffold_command.params.append(
                click.Option(
                    ["--json-params"],
                    type=str,
                    default=None,
                    help="JSON string of scaffolder parameters. Mutually exclusive with passing individual parameters as options.",
                    callback=parse_json_option,
                )
            )
            for name, field_info in obj.scaffolder_schema["properties"].items():
                # All fields are currently optional because they can also be passed under
                # `--json-params`
                option = json_schema_property_to_click_option(name, field_info, required=False)
                scaffold_command.params.append(option)

    def _get_matching_command(self, ctx: click.Context, input_cmd: str) -> click.Command:
        commands = self.list_commands(ctx)
        cmd_query = input_cmd.lower()
        matches = sorted([name for name in commands if cmd_query in name.lower()])

        # if input is not a substring match for any registered command, try to interpret it as a
        # Python reference, load the corresponding registry object, and generate a command on the
        # fly
        if len(matches) == 0:
            snap = self._try_load_input_as_registry_object(input_cmd)
            if snap:
                self._create_subcommand(snap.key, snap, use_typename_alias=False)
                return check.not_none(super().get_command(ctx, snap.key.to_typename()))
            else:
                exit_with_error(generate_missing_registry_object_error_message(input_cmd))

        if len(matches) == 1:
            click.echo(f"No exact match found for '{input_cmd}'. Did you mean this one?")
            click.echo(f"    {matches[0]}")
            selection = click.prompt("Choose (y/n)", type=str, default="y")
            if selection == "y":
                index = 1
            elif selection == "n":
                click.echo("Exiting.")
                ctx.exit(0)
            else:
                exit_with_error(f"Invalid selection: {selection}. Please choose 'y' or 'n'.")
        else:
            # Present a menu of options for the user to choose from
            click.echo(f"No exact match found for '{input_cmd}'. Did you mean one of these?")
            for i, match in enumerate(matches, 1):
                click.echo(f"({i}) {match}")
            click.echo("(n) quit")

            # Get user selection
            selection = click.prompt("Select an option (number)", type=str, default="1")
            if selection == "n":
                click.echo("Exiting.")
                ctx.exit(0)

            invalid_selection_msg = f"Invalid selection: {selection}. Please choose a number between 1 and {len(matches)}."
            if not selection.isdigit():
                exit_with_error(invalid_selection_msg)
            index = int(selection)
            if index < 1 or index > len(matches):
                exit_with_error(invalid_selection_msg)

        selected_cmd = matches[index - 1]
        click.echo(f"Using defs scaffolder: {selected_cmd}")
        return check.not_none(super().get_command(ctx, selected_cmd))

    def _try_load_input_as_registry_object(self, input_str: str) -> EnvRegistryObjectSnap | None:
        from dagster.components.core.snapshot import get_package_entry_snap

        if not EnvRegistryKey.is_valid_typename(input_str):
            return None
        key = EnvRegistryKey.from_typename(input_str)
        try:
            obj = load_module_object(key.namespace, key.name)
            return get_package_entry_snap(key, obj)
        except DagsterUnresolvableSymbolError:
            return None


class ScaffoldDefsSubCommand(DgClickCommand):
    # We have to override this because the implementation of `format_help` used elsewhere will only
    # pull parameters directly off the target command. For these component scaffold subcommands  we need
    # to expose the global options, which are defined on the preceding group rather than the command
    # itself.
    def format_help(self, context: click.Context, formatter: click.HelpFormatter):
        """Customizes the help to include hierarchical usage."""
        from typer.rich_utils import rich_format_help

        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")

        # This is a hack. We pass the help format func a modified version of the command where the global
        # options are attached to the command itself. This will cause them to be included in the
        # help output.
        cmd_copy = copy(self)
        cmd_copy.params = [
            *cmd_copy.params,
            *(GLOBAL_OPTIONS.values()),
        ]
        rich_format_help(obj=cmd_copy, ctx=context, markup_mode="rich")

    def format_usage(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")
        arg_pieces = self.collect_usage_pieces(ctx)
        command_parts = ctx.command_path.split(" ")
        command_parts.insert(-1, "[GLOBAL OPTIONS]")
        return formatter.write_usage(" ".join(command_parts), " ".join(arg_pieces))


# We have to override the usual Click processing of `--help` here. The issue is
# that click will process this option before processing anything else, but because we are
# dynamically generating subcommands based on the content of other options, the output of --help
# actually depends on these other options. So we opt out of Click's short-circuiting
# behavior of `--help` by setting `help_option_names=[]`, ensuring that we can process the other
# options first and generate the correct subcommands. We then add a custom `--help` option that
# gets invoked inside the callback.
@click.group(
    name="defs",
    cls=ScaffoldDefsGroup,
    invoke_without_command=True,
    context_settings={"help_option_names": []},
)
@click.option("-h", "--help", "help_", is_flag=True, help="Show this message and exit.")
@dg_global_options
@click.pass_context
def scaffold_defs_group(context: click.Context, help_: bool, **global_options: object) -> None:
    """Commands for scaffolding Dagster code."""
    # Click attempts to resolve subcommands BEFORE it invokes this callback.
    # Therefore we need to manually invoke this callback during subcommand generation to make sure
    # it runs first. It will be invoked again later by Click. We make it idempotent to deal with
    # that.
    if not has_config_on_cli_context(context):
        cli_config = normalize_cli_config(global_options, context)
        set_config_on_cli_context(context, cli_config)
    if help_:
        click.echo(context.get_help())
        context.exit(0)


@scaffold_defs_group.command(
    name="inline-component",
    cls=ScaffoldDefsSubCommand,
)
@click.argument("path", type=str)
@click.option(
    "--typename",
    type=str,
    required=True,
)
@click.option(
    "--superclass",
    type=str,
    help="The superclass for the component to scaffold. If unset, `dg.Component` will be used.",
)
@cli_telemetry_wrapper
def scaffold_defs_inline_component(
    path: str,
    typename: str,
    superclass: str | None,
) -> None:
    """Scaffold a new Dagster component."""
    # We need to pass the global options to the command, but we don't want to
    # pass them to the subcommand. So we remove them from the context.
    context = click.get_current_context()
    cli_config = get_config_from_cli_context(context)
    dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)

    if dg_context.has_object_at_defs_path(path):
        exit_with_error(f"A component instance at `{path}` already exists.")

    scaffold_inline_component(
        Path(path),
        typename,
        superclass,
        dg_context,
    )


def _core_scaffold(
    dg_context: DgContext,
    cli_context: click.Context,
    object_key: EnvRegistryKey,
    defs_path: str,
    key_value_params: Mapping[str, Any],
    scaffold_format: ScaffoldFormatOptions,
    json_params: Mapping[str, Any] | None = None,
) -> None:
    from dagster.components.core.package_entry import is_scaffoldable_object_key
    from pydantic import ValidationError

    if not is_scaffoldable_object_key(object_key):
        exit_with_error(f"Scaffoldable object type `{object_key.to_typename()}` not found.")
    elif dg_context.has_object_at_defs_path(defs_path):
        exit_with_error(f"Path `{(dg_context.defs_path / defs_path).absolute()}` already exists.")

    # Specified key-value params will be passed to this function with their default value of
    # `None` even if the user did not set them. Filter down to just the ones that were set by
    # the user.
    user_provided_key_value_params = {
        k: v
        for k, v in key_value_params.items()
        if cli_context.get_parameter_source(k) == ParameterSource.COMMANDLINE
    }
    if json_params is not None and user_provided_key_value_params:
        exit_with_error(
            "Detected params passed as both --json-params and individual options. These are mutually exclusive means of passing"
            " component generation parameters. Use only one.",
        )
    elif json_params:
        scaffold_params = json_params
    elif user_provided_key_value_params:
        scaffold_params = user_provided_key_value_params
    else:
        scaffold_params = None

    try:
        scaffold_registry_object(
            Path(dg_context.defs_path) / defs_path,
            object_key.to_typename(),
            scaffold_params,
            dg_context,
            scaffold_format,
        )
    except ValidationError as e:
        exit_with_error(
            (
                f"Error validating scaffold parameters for `{object_key.to_typename()}`:\\n\\n"
                f"{e.json(indent=4)}"
            ),
            do_format=False,
        )



"""Test run event business logic functions and CLI command invocation.

Includes a regression test for:
    TypeError: get_run_events_command() got an unexpected keyword argument 'view_graphql'
"""

import json

from click.testing import CliRunner
from dagster_dg_cli.api_layer.schemas.run_event import DgApiRunEvent, RunEventLevel, RunEventList
from dagster_dg_cli.cli.api.client import DgApiTestContext
from dagster_dg_cli.cli.api.run_event import format_run_events_json, format_run_events_table

_RUN_ID = "a34cd75c-cfa9-4e99-8f0a-955492b89afd"

_SAMPLE_GRAPHQL_RESPONSE = {
    "logsForRun": {
        "__typename": "EventConnection",
        "events": [
            {
                "__typename": "MessageEvent",
                "runId": _RUN_ID,
                "message": 'Started execution of run for "dbt_analytics_core_job".',
                "timestamp": "1771476078297",
                "level": "INFO",
                "stepKey": None,
                "eventType": "PIPELINE_START",
            },
            {
                "__typename": "MessageEvent",
                "runId": _RUN_ID,
                "message": 'dbt_analytics_core_job intends to materialize asset "my_asset".',
                "timestamp": "1771476078298",
                "level": "INFO",
                "stepKey": "dbt_non_partitioned_models_2",
                "eventType": "ASSET_MATERIALIZATION_PLANNED",
            },
        ],
        "cursor": None,
        "hasMore": False,
    }
}


def _make_sample_events() -> RunEventList:
    return RunEventList(
        items=[
            DgApiRunEvent(
                run_id=_RUN_ID,
                message='Started execution of run for "dbt_analytics_core_job".',
                timestamp="1771476078297",
                level=RunEventLevel.INFO,
                step_key=None,
                event_type="PIPELINE_START",
            ),
            DgApiRunEvent(
                run_id=_RUN_ID,
                message='dbt_analytics_core_job intends to materialize asset "my_asset".',
                timestamp="1771476078298",
                level=RunEventLevel.INFO,
                step_key="dbt_non_partitioned_models_2",
                event_type="ASSET_MATERIALIZATION_PLANNED",
            ),
        ],
        total=2,
    )


class TestFormatRunEventsTable:
    def test_format_with_events(self, snapshot):
        events = _make_sample_events()
        result = format_run_events_table(events, _RUN_ID)
        snapshot.assert_match(result)

    def test_format_empty_events(self):
        events = RunEventList(items=[], total=0)
        result = format_run_events_table(events, _RUN_ID)
        assert result == f"No events found for run {_RUN_ID}"


class TestFormatRunEventsJson:
    def test_format_with_events(self, snapshot):
        events = _make_sample_events()
        result = format_run_events_json(events, _RUN_ID)
        snapshot.assert_match(json.loads(result))

    def test_format_empty_events(self):
        events = RunEventList(items=[], total=0)
        result = format_run_events_json(events, _RUN_ID)
        parsed = json.loads(result)
        assert parsed["run_id"] == _RUN_ID
        assert parsed["events"] == []
        assert parsed["count"] == 0


class TestGetRunEventsCommandInvocation:
    """Regression tests for `dg api run-events get` CLI invocation."""

    def test_command_accepts_view_graphql_param(self):
        """Regression test: get_run_events_command must accept the view_graphql kwarg.

        The @dg_api_options decorator unconditionally injects view_graphql into kwargs
        (dagster_shared/plus/config_utils.py). If the command function signature is missing
        this parameter, the invocation crashes with:

            TypeError: get_run_events_command() got an unexpected keyword argument 'view_graphql'
        """
        from dagster_dg_cli.cli import cli as root_cli

        from dagster_dg_cli_tests.cli_tests.api_tests.test_dynamic_command_execution import (
            ReplayClient,
        )

        replay_client = ReplayClient([_SAMPLE_GRAPHQL_RESPONSE])
        test_context = DgApiTestContext(client_factory=lambda config: replay_client)

        runner = CliRunner()
        result = runner.invoke(root_cli, ["api", "run-events", "get", _RUN_ID], obj=test_context)

        assert "view_graphql" not in (result.output or ""), (
            "Command crashed with view_graphql TypeError - check get_run_events_command signature"
        )
        assert result.exit_code == 0



from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dagster._config import Field, Shape
from dagster._core.instance.types import T_DagsterInstance
from dagster._core.storage.defs_state.base import DefsStateStorage
from dagster._serdes.config_class import ConfigurableClass, ConfigurableClassData
from dagster_cloud_cli.core.artifacts import download_artifact, upload_artifact
from dagster_cloud_cli.core.headers.auth import DagsterCloudInstanceScope
from dagster_shared import check
from dagster_shared.serdes.objects.models.defs_state_info import DefsStateInfo

if TYPE_CHECKING:
    from dagster_cloud_cli.commands.ci.state import LocationState


GET_LATEST_DEFS_STATE_INFO_QUERY = """
    query getLatestDefsStateInfo {
        latestDefsStateInfo {
            keyStateInfo {
                name
                info {
                    version
                    createTimestamp
                }
            }
        }
    }
"""

SET_LATEST_VERSION_MUTATION = """
    mutation setLatestDefsStateVersion($key: String!, $version: String!) {
        setLatestDefsStateVersion(key: $key, version: $version) {
            ok
        }
    }
"""


class DagsterPlusCliDefsStateStorage(DefsStateStorage[T_DagsterInstance], ConfigurableClass):
    """DefsStateStorage that can be instantiated from a DagsterPlusCliConfig,
    intended for use within the CLI.
    """

    def __init__(
        self,
        url: str,
        api_token: str,
        deployment: str,
        graphql_client,
        organization: str,
        inst_data: ConfigurableClassData | None = None,
    ):
        self._url = url
        self._api_token = api_token
        self._deployment = deployment
        self._graphql_client = graphql_client
        self._organization = organization
        self._inst_data = inst_data

    @classmethod
    def from_location_state(
        cls, location_state: "LocationState", api_token: str, organization: str
    ):
        from dagster_dg_cli.utils.plus.gql_client import DagsterPlusGraphQLClient

        return cls(
            location_state.url,
            api_token,
            location_state.deployment_name,
            DagsterPlusGraphQLClient.from_location_state(location_state, api_token, organization),
            organization,
        )

    @property
    def inst_data(self) -> ConfigurableClassData | None:
        return self._inst_data

    @classmethod
    def config_type(cls):
        return Shape(
            {
                "url": Field(str, is_required=True),
                "api_token": Field(str, is_required=True),
                "deployment": Field(str, is_required=True),
                "organization": Field(str, is_required=True),
            }
        )

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Mapping[str, Any]):
        from dagster_dg_cli.utils.plus.gql_client import DagsterPlusGraphQLClient

        url = check.str_param(config_value.get("url"), "url")
        api_token = check.str_param(config_value.get("api_token"), "api_token")
        deployment = check.str_param(config_value.get("deployment"), "deployment")
        organization = check.str_param(config_value.get("organization"), "organization")

        # Create graphql client from config
        graphql_client = DagsterPlusGraphQLClient(
            url=f"{url}/graphql",
            headers={
                "Dagster-Cloud-Api-Token": api_token,
                "Dagster-Cloud-Organization": organization,
                "Dagster-Cloud-Deployment": deployment,
            },
        )

        return cls(
            url=url,
            api_token=api_token,
            deployment=deployment,
            graphql_client=graphql_client,
            organization=organization,
            inst_data=inst_data,
        )

    @property
    def url(self) -> str:
        return self._url

    @property
    def api_token(self) -> str:
        return self._api_token

    @property
    def deployment(self) -> str:
        return self._deployment

    @property
    def graphql_client(self) -> Any:
        return self._graphql_client

    def _execute_query(self, query, variables=None):
        return self.graphql_client.execute(query, variables=variables)

    def _get_artifact_key(self, key: str, version: str) -> str:
        return f"__state__/{self._sanitize_key(key)}/{version}"

    def download_state_to_path(self, key: str, version: str, path: Path) -> None:
        download_artifact(
            url=self.url,
            scope=DagsterCloudInstanceScope.DEPLOYMENT,
            api_token=self.api_token,
            key=self._get_artifact_key(key, version),
            path=path,
            deployment=self.deployment,
        )

    def upload_state_from_path(self, key: str, version: str, path: Path) -> None:
        upload_artifact(
            url=self.url,
            scope=DagsterCloudInstanceScope.DEPLOYMENT,
            api_token=self.api_token,
            key=self._get_artifact_key(key, version),
            path=path,
            deployment=self.deployment,
        )
        self.set_latest_version(key, version)

    def get_latest_defs_state_info(self) -> DefsStateInfo | None:
        res = self._execute_query(GET_LATEST_DEFS_STATE_INFO_QUERY)
        latest_info = res["latestDefsStateInfo"]
        return DefsStateInfo.from_graphql(latest_info) if latest_info else None

    def set_latest_version(self, key: str, version: str) -> None:
        result = self._execute_query(
            SET_LATEST_VERSION_MUTATION, variables={"key": key, "version": version}
        )
        check.invariant(
            result.get("setLatestDefsStateVersion", {}).get("ok"),
            f"Failed to set latest version. Result: {result}",
        )



"""Legacy command for scaffolding GitHub Actions workflow.

This command is maintained for backward compatibility.
Consider using `dg plus deploy configure [serverless|hybrid] --git-provider github` instead.
"""

from pathlib import Path

import click
from dagster_dg_core.config import normalize_cli_config
from dagster_dg_core.context import DgContext
from dagster_dg_core.shared_options import dg_global_options
from dagster_dg_core.utils import DgClickCommand
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper

from dagster_dg_cli.cli.plus.constants import DgPlusAgentType
from dagster_dg_cli.cli.plus.deploy.configure.commands import (
    resolve_deployment,
    resolve_git_root,
    resolve_organization,
    resolve_python_version,
    resolve_url,
)
from dagster_dg_cli.cli.plus.deploy.configure.configure_ci import configure_ci_impl
from dagster_dg_cli.cli.plus.deploy.configure.utils import DgPlusDeployConfigureOptions, GitProvider
from dagster_dg_cli.utils.plus.build import get_agent_type_and_platform_from_graphql
from dagster_dg_cli.utils.plus.gql_client import DagsterPlusGraphQLClient


def _resolve_config_for_github_actions(
    git_root: Path | None,
    dg_context: DgContext,
    cli_config,
) -> DgPlusDeployConfigureOptions:
    """Resolve config for legacy github-actions command.

    This command prompts in the legacy order: org, deployment, agent type (no platform).
    """
    from dagster_shared.plus.config import DagsterPlusCliConfig

    plus_config = DagsterPlusCliConfig.get() if DagsterPlusCliConfig.exists() else None

    # Prompt for org, URL, and deployment FIRST (legacy order)
    resolved_organization = resolve_organization(None, plus_config)
    resolved_url = resolve_url(None, plus_config, resolved_organization)
    resolved_deployment = resolve_deployment(None, plus_config)

    # Try to detect agent type and platform from Plus config
    agent_type = None
    agent_platform = None
    if plus_config:
        try:
            gql_client = DagsterPlusGraphQLClient.from_config(plus_config)
            agent_type, agent_platform = get_agent_type_and_platform_from_graphql(gql_client)
        except Exception:
            pass

    # Prompt for agent type if not detected
    if agent_type is None:
        agent_type = DgPlusAgentType(
            click.prompt(
                "Deployment agent type",
                type=click.Choice(["serverless", "hybrid"]),
            ).upper()
        )

    # Resolve git root
    resolved_git_root = resolve_git_root(git_root, GitProvider.GITHUB)

    # Use default Python version
    resolved_python_version = resolve_python_version(None)

    return DgPlusDeployConfigureOptions(
        dg_context=dg_context,
        cli_config=cli_config,
        plus_config=plus_config,
        agent_type=agent_type,
        agent_platform=agent_platform,  # May be None - legacy didn't require platform
        organization_name=resolved_organization,
        cloud_url=resolved_url,
        deployment_name=resolved_deployment,
        git_root=resolved_git_root,
        python_version=resolved_python_version,
        skip_confirmation_prompt=False,
        git_provider=GitProvider.GITHUB,
        use_editable_dagster=False,
    )


@click.command(
    name="github-actions",
    cls=DgClickCommand,
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.option("--git-root", type=Path, help="Path to the git root of the repository")
@dg_global_options
@cli_telemetry_wrapper
def scaffold_github_actions_command(git_root: Path | None, **global_options: object) -> None:
    """Scaffold a GitHub Actions workflow for a Dagster project.

    This command will create a GitHub Actions workflow in the `.github/workflows` directory.

    NOTE: This command is maintained for backward compatibility.
    Consider using `dg plus deploy configure [serverless|hybrid] --git-provider github`
    instead for a complete deployment setup.
    """
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_workspace_or_project_environment(Path.cwd(), cli_config)

    config = _resolve_config_for_github_actions(
        git_root=git_root,
        dg_context=dg_context,
        cli_config=cli_config,
    )

    configure_ci_impl(config)



import shutil
import signal
import tempfile
import textwrap
from pathlib import Path

import pytest
from dagster_dg_core.context import DG_PROJECT_PYTHON_EXECUTABLE_ENV_VAR
from dagster_dg_core.utils import activate_venv, discover_repo_root, is_windows, pushd
from dagster_shared.utils import environ
from dagster_test.components.test_utils.test_cases import BASIC_INVALID_VALUE, BASIC_MISSING_VALUE
from dagster_test.dg_utils.utils import (
    ProxyRunner,
    assert_projects_loaded_and_exit,
    assert_runner_result,
    create_project_from_components,
    find_free_port,
    install_editable_dg_dev_packages_to_venv,
    isolated_example_project_foo_bar,
    isolated_example_workspace,
    launch_dev_command,
)


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_dev_workspace_context_success(monkeypatch):
    # The command will use `uv tool run dagster dev` to start the webserver if it
    # cannot find a venv with `dagster` and `dagster-webserver` installed. `uv tool run` will
    # pull the `dagster` package from PyPI. To avoid this, we ensure the workspace directory has a
    # venv with `dagster` and `dagster-webserver` installed.
    dagster_git_repo_dir = str(discover_repo_root(Path(__file__)))
    with (
        ProxyRunner.test() as runner,
        isolated_example_workspace(runner, create_venv=True) as workspace_path,
        environ({"DAGSTER_GIT_REPO_DIR": dagster_git_repo_dir}),
    ):
        with activate_venv(workspace_path / ".venv"):
            result = runner.invoke_create_dagster(
                "project",
                "--use-editable-dagster",
                "project-1",
                "--uv-sync",
            )
            assert_runner_result(result)
            result = runner.invoke_create_dagster(
                "project",
                "--use-editable-dagster",
                "project-2",
                "--uv-sync",
            )
            assert_runner_result(result)

            (Path("project-2") / "src" / "project_2" / "defs" / "my_asset.py").write_text(
                "import dagster as dg\n\n@dg.asset\ndef my_asset(): pass"
            )
            port = find_free_port()
            dev_process = launch_dev_command(["--port", str(port)])
            projects = {"project-1", "project-2"}
            assert_projects_loaded_and_exit(projects, port, dev_process)


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_dev_workspace_context_set_python_executable_from_env_file():
    """Test that the dg dev command properly loads env files from the workspace and projects."""
    dagster_git_repo_dir = str(discover_repo_root(Path(__file__)))
    with (
        ProxyRunner.test() as runner,
        isolated_example_workspace(runner, create_venv=True) as workspace_path,
        environ({"DAGSTER_GIT_REPO_DIR": dagster_git_repo_dir}),
    ):
        with activate_venv(workspace_path / ".venv"):
            result = runner.invoke_create_dagster(
                "project",
                "--use-editable-dagster",
                "project-1",
                "--uv-sync",
            )
            assert_runner_result(result)

            # Now we move the venv to a different location to test
            # DG_PROJECT_PYTHON_EXECUTABLE_ENV_VAR is used.
            Path("project-1/.env").write_text(
                f"{DG_PROJECT_PYTHON_EXECUTABLE_ENV_VAR}=../._venv/bin/python\n"
            )
            shutil.move("project-1/.venv", "._venv")

            result = runner.invoke_create_dagster(
                "project",
                "--use-editable-dagster",
                "project-2",
                "--uv-sync",
            )
            assert_runner_result(result)

            # test with quoted value
            Path("project-2/.env").write_text(
                f"{DG_PROJECT_PYTHON_EXECUTABLE_ENV_VAR}=._venv/bin/python\n"
            )
            shutil.move("project-2/.venv", "project-2/._venv")

            port = find_free_port()
            with (
                tempfile.NamedTemporaryFile() as stdout_file,
                open(stdout_file.name, "w") as stdout,
            ):
                dev_process = launch_dev_command(["--port", str(port)], stdout=stdout)
                projects = {"project-1", "project-2"}
                assert_projects_loaded_and_exit(projects, port, dev_process)

                assert ("Environment variables will not be injected") not in Path(
                    stdout_file.name
                ).read_text()


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_dev_workspace_load_env_files(monkeypatch):
    """Test that the dg dev command properly loads env files from the workspace and projects."""
    dagster_git_repo_dir = str(discover_repo_root(Path(__file__)))
    with (
        ProxyRunner.test() as runner,
        isolated_example_workspace(runner, create_venv=True) as workspace_path,
        environ({"DAGSTER_GIT_REPO_DIR": dagster_git_repo_dir}),
    ):
        with activate_venv(workspace_path / ".venv"):
            Path(".env").write_text("WORKSPACE_ENV_VAR=1\nOVERWRITTEN_ENV_VAR=3")
            result = runner.invoke_create_dagster(
                "project",
                "--use-editable-dagster",
                "project-1",
                "--uv-sync",
            )
            assert_runner_result(result)
            result = runner.invoke_create_dagster(
                "project",
                "--use-editable-dagster",
                "project-2",
                "--uv-sync",
            )
            assert_runner_result(result)

            (Path("project-2") / ".env").write_text("PROJECT_ENV_VAR=2\nOVERWRITTEN_ENV_VAR=4")

            (Path("project-2") / "src" / "project_2" / "defs" / "my_def.py").write_text(
                textwrap.dedent("""
                import os

                assert os.environ["PROJECT_ENV_VAR"] == "2"
                assert os.environ["WORKSPACE_ENV_VAR"] == "1"
                assert os.environ["OVERWRITTEN_ENV_VAR"] == "4"
                """).strip()
            )
            port = find_free_port()
            with (
                tempfile.NamedTemporaryFile() as stdout_file,
                open(stdout_file.name, "w") as stdout,
            ):
                dev_process = launch_dev_command(["--port", str(port)], stdout=stdout)
                projects = {"project-1", "project-2"}
                assert_projects_loaded_and_exit(projects, port, dev_process)

                assert ("Environment variables will not be injected") not in Path(
                    stdout_file.name
                ).read_text()


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_dev_project_context_success():
    with (
        ProxyRunner.test() as runner,
        isolated_example_project_foo_bar(runner, uv_sync=True) as project_path,
    ):
        venv_path = project_path / ".venv"
        install_editable_dg_dev_packages_to_venv(venv_path)

        with activate_venv(venv_path):
            port = find_free_port()
            dev_process = launch_dev_command(["--port", str(port)])
            assert_projects_loaded_and_exit({"foo-bar"}, port, dev_process)


@pytest.mark.skipif(
    is_windows() == "Windows", reason="Temporarily skipping (signal issues in CLI).."
)
def test_dev_has_options_of_dagster_dev():
    from dagster._cli.dev import dev_command as dagster_dev_command
    from dagster_dg_cli.cli import dev_command as dev_command

    exclude_dagster_dev_params = {
        # Exclude options that are used to set the target. `dg dev` does not use.
        "empty_workspace",
        "workspace",
        "python_file",
        "module_name",
        "package_name",
        "attribute",
        "working_directory",
        "grpc_port",
        "grpc_socket",
        "grpc_host",
        "use_ssl",
        # Misc others to exclude
        "use_legacy_code_server_behavior",
        "shutdown_pipe",
    }

    dg_dev_param_names = {param.name for param in dev_command.params}
    dagster_dev_param_names = {param.name for param in dagster_dev_command.params}
    dagster_dev_params_to_check = dagster_dev_param_names - exclude_dagster_dev_params

    unmatched_params = dagster_dev_params_to_check - dg_dev_param_names
    assert not unmatched_params, f"dg dev missing params: {unmatched_params}"


def test_implicit_yaml_check_from_dg_dev() -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_MISSING_VALUE.component_path,
            BASIC_INVALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_MISSING_VALUE.component_type_filepath,
            in_workspace=False,
        ) as tmpdir,
    ):
        with pushd(str(tmpdir)):
            result = runner.invoke("dev")
            assert result.exit_code != 0, str(result.output)

            assert BASIC_INVALID_VALUE.check_error_msg and BASIC_MISSING_VALUE.check_error_msg
            BASIC_INVALID_VALUE.check_error_msg(str(result.output))
            BASIC_MISSING_VALUE.check_error_msg(str(result.output))


def test_implicit_yaml_check_from_dg_dev_in_workspace_context() -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_MISSING_VALUE.component_path,
            BASIC_INVALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_MISSING_VALUE.component_type_filepath,
            in_workspace=True,
        ) as tmpdir,
    ):
        with pushd(Path(tmpdir).parent):
            result = runner.invoke("dev", "--check-yaml")
            assert result.exit_code != 0, str(result.output)

            assert "--check-yaml is not currently supported in a workspace context" in str(
                result.output
            )

        # It is supported and is the default in a project context within a workspace
        with pushd(tmpdir):
            result = runner.invoke("dev", "--check-yaml")
            assert result.exit_code != 0, str(result.output)

            assert BASIC_INVALID_VALUE.check_error_msg and BASIC_MISSING_VALUE.check_error_msg
            BASIC_INVALID_VALUE.check_error_msg(str(result.output))
            BASIC_MISSING_VALUE.check_error_msg(str(result.output))


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_dev_uses_active_venv_when_flag_set():
    """Test that dev command logs the active venv Python when --use-active-venv is set."""
    dagster_git_repo_dir = str(discover_repo_root(Path(__file__)))
    with (
        ProxyRunner.test() as runner,
        isolated_example_workspace(runner, create_venv=True) as workspace_path,
        environ({"DAGSTER_GIT_REPO_DIR": dagster_git_repo_dir}),
    ):
        venv_path = workspace_path / ".venv"
        install_editable_dg_dev_packages_to_venv(venv_path)

        with activate_venv(venv_path):
            # Create a project
            result = runner.invoke_create_dagster(
                "project", "--use-editable-dagster", "test-project", "--uv-sync"
            )
            assert_runner_result(result)

            # Start dev server with --use-active-venv flag and capture output
            port = find_free_port()
            with (
                tempfile.NamedTemporaryFile(mode="w+") as stdout_file,
                tempfile.NamedTemporaryFile(mode="w+") as stderr_file,
            ):
                with open(stdout_file.name, "w") as stdout, open(stderr_file.name, "w") as stderr:
                    dev_process = launch_dev_command(
                        ["--port", str(port), "--use-active-venv"], stdout=stdout, stderr=stderr
                    )

                # Give it a moment to start and log the message
                import time

                time.sleep(2)

                # Read the captured output
                with open(stdout_file.name) as f:
                    stdout_content = f.read()
                with open(stderr_file.name) as f:
                    stderr_content = f.read()

                # Clean up the process
                dev_process.send_signal(signal.SIGINT)
                dev_process.communicate()

                # Verify the log message is present
                combined_output = stdout_content + stderr_content
                assert "Using active Python environment:" in combined_output, (
                    f"Expected log message about using active Python environment, but got:\n{combined_output}"
                )



"""Registry of all API classes for REST compliance testing."""

from dagster_dg_cli.api_layer.api.asset import DgApiAssetApi
from dagster_dg_cli.api_layer.api.deployments import DgApiDeploymentApi


def get_all_api_classes():
    """Get all API classes to test."""
    return [
        DgApiDeploymentApi,
        DgApiAssetApi,
    ]



import click
from dagster_dg_core.utils import DgClickGroup

from dagster_dg_cli.cli.plus.create.ci_api_token import create_ci_api_token
from dagster_dg_cli.cli.plus.create.env import create_env_command


@click.group(name="create", cls=DgClickGroup)
def plus_create_group():
    """Commands for creating configuration in Dagster Plus."""


plus_create_group.add_command(create_env_command)
plus_create_group.add_command(create_ci_api_token)



"""Test sensor business logic functions without mocks.

These tests focus on testing pure functions that process data without requiring
GraphQL client mocking or external dependencies.
"""

import json

from dagster_dg_cli.api_layer.graphql_adapter.sensor import (
    process_repositories_response,
    process_sensor_response,
    process_sensors_response,
)
from dagster_dg_cli.api_layer.schemas.sensor import (
    DgApiSensor,
    DgApiSensorList,
    DgApiSensorStatus,
    DgApiSensorType,
)
from dagster_dg_cli.cli.api.sensor import format_sensor, format_sensors


class TestProcessSensorResponses:
    """Test the pure functions that process GraphQL responses."""

    def test_process_sensors_response_success(self, snapshot):
        """Test processing a successful sensors GraphQL response."""
        # Sample GraphQL response structure
        response = {
            "sensorsOrError": {
                "__typename": "Sensors",
                "results": [
                    {
                        "id": "sensor1-id",
                        "name": "daily_sensor",
                        "sensorState": {"status": "RUNNING"},
                        "sensorType": "STANDARD",
                        "description": "Daily processing sensor",
                    },
                    {
                        "id": "sensor2-id",
                        "name": "asset_sensor",
                        "sensorState": {"status": "STOPPED"},
                        "sensorType": "ASSET",
                        "description": "Asset change sensor",
                    },
                ],
            }
        }
        result = process_sensors_response(response)

        # Snapshot the entire result to capture structure and data
        snapshot.assert_match(result)

    def test_process_sensors_response_empty(self, snapshot):
        """Test processing an empty sensors GraphQL response."""
        response = {"sensorsOrError": {"__typename": "Sensors", "results": []}}
        result = process_sensors_response(response)

        # Snapshot empty result
        snapshot.assert_match(result)

    def test_process_sensors_response_missing_key(self, snapshot):
        """Test processing a response missing the sensorsOrError key."""
        malformed_response = {}
        try:
            result = process_sensors_response(malformed_response)
            # If no exception, snapshot the result
            snapshot.assert_match(result)
        except Exception as e:
            # Snapshot the error message
            snapshot.assert_match({"error": str(e)})

    def test_process_sensors_response_error_typename(self, snapshot):
        """Test processing a sensors response with error typename."""
        error_response = {
            "sensorsOrError": {
                "__typename": "RepositoryNotFoundError",
                "message": "Repository not found",
            }
        }
        try:
            result = process_sensors_response(error_response)
            snapshot.assert_match(result)
        except Exception as e:
            snapshot.assert_match({"error": str(e)})

    def test_process_repositories_response_success(self, snapshot):
        """Test processing a successful repositories GraphQL response."""
        response = {
            "repositoriesOrError": {
                "__typename": "RepositoryConnection",
                "nodes": [
                    {
                        "name": "main_repo",
                        "location": {"name": "main_location"},
                        "sensors": [
                            {
                                "id": "sensor1-id",
                                "name": "daily_sensor",
                                "sensorState": {"status": "RUNNING"},
                                "sensorType": "STANDARD",
                                "description": "Daily processing sensor",
                            },
                            {
                                "id": "sensor2-id",
                                "name": "asset_sensor",
                                "sensorState": {"status": "PAUSED"},
                                "sensorType": "MULTI_ASSET",
                                "description": "Asset change sensor",
                            },
                        ],
                    },
                    {
                        "name": "secondary_repo",
                        "location": {"name": "secondary_location"},
                        "sensors": [
                            {
                                "id": "sensor3-id",
                                "name": "hourly_sensor",
                                "sensorState": {"status": "STOPPED"},
                                "sensorType": "FRESHNESS_POLICY",
                                "description": None,
                            }
                        ],
                    },
                ],
            }
        }
        result = process_repositories_response(response)

        # Snapshot the entire result to capture structure and data
        snapshot.assert_match(result)

    def test_process_repositories_response_empty(self, snapshot):
        """Test processing an empty repositories GraphQL response."""
        response = {"repositoriesOrError": {"__typename": "RepositoryConnection", "nodes": []}}
        result = process_repositories_response(response)

        snapshot.assert_match(result)

    def test_process_sensor_response_success(self, snapshot):
        """Test processing a successful single sensor GraphQL response."""
        response = {
            "sensorOrError": {
                "__typename": "Sensor",
                "id": "single-sensor-id",
                "name": "critical_sensor",
                "sensorState": {"status": "RUNNING"},
                "sensorType": "AUTO_MATERIALIZE",
                "description": "Critical production sensor",
            }
        }
        result = process_sensor_response(response)

        snapshot.assert_match(result)

    def test_process_sensor_response_not_found_error(self, snapshot):
        """Test processing a sensor not found error."""
        error_response = {
            "sensorOrError": {
                "__typename": "SensorNotFoundError",
                "message": "Sensor 'nonexistent_sensor' not found",
            }
        }
        try:
            result = process_sensor_response(error_response)
            snapshot.assert_match(result)
        except Exception as e:
            snapshot.assert_match({"error": str(e)})


class TestFormatSensors:
    """Test the sensor formatting functions."""

    def _create_sample_sensor_list(self):
        """Create sample DgApiSensorList for testing."""
        sensors = [
            DgApiSensor(
                id="sensor1-id",
                name="daily_sensor",
                status=DgApiSensorStatus.RUNNING,
                sensor_type=DgApiSensorType.STANDARD,
                description="Daily processing sensor",
                repository_origin="main_location@main_repo",
                next_tick_timestamp=1705311000.0,  # 2024-01-15T10:30:00Z
            ),
            DgApiSensor(
                id="sensor2-id",
                name="asset_sensor",
                status=DgApiSensorStatus.STOPPED,
                sensor_type=DgApiSensorType.ASSET,
                description="Asset change sensor",
                repository_origin="main_location@main_repo",
                next_tick_timestamp=None,
            ),
            DgApiSensor(
                id="sensor3-id",
                name="minimal_sensor",
                status=DgApiSensorStatus.PAUSED,
                sensor_type=DgApiSensorType.MULTI_ASSET,
                description=None,
                repository_origin=None,
                next_tick_timestamp=None,
            ),
        ]
        return DgApiSensorList(items=sensors, total=len(sensors))

    def _create_sample_sensor(self):
        """Create sample DgApiSensor for testing."""
        return DgApiSensor(
            id="single-sensor-id",
            name="critical_sensor",
            status=DgApiSensorStatus.RUNNING,
            sensor_type=DgApiSensorType.AUTO_MATERIALIZE,
            description="Critical production sensor",
            repository_origin="prod_location@prod_repo",
            next_tick_timestamp=1705311900.0,  # 2024-01-15T10:45:00Z
        )

    def test_format_sensors_text_output(self, snapshot):
        """Test formatting sensors as text."""
        from dagster_shared.utils.timing import fixed_timezone

        sensor_list = self._create_sample_sensor_list()
        with fixed_timezone("UTC"):
            result = format_sensors(sensor_list, as_json=False)

        # Snapshot the entire text output
        snapshot.assert_match(result)

    def test_format_sensors_json_output(self, snapshot):
        """Test formatting sensors as JSON."""
        sensor_list = self._create_sample_sensor_list()
        result = format_sensors(sensor_list, as_json=True)

        # For JSON, we want to snapshot the parsed structure to avoid formatting differences
        parsed = json.loads(result)
        snapshot.assert_match(parsed)

    def test_format_sensor_text_output(self, snapshot):
        """Test formatting single sensor as text."""
        from dagster_shared.utils.timing import fixed_timezone

        sensor = self._create_sample_sensor()
        with fixed_timezone("UTC"):
            result = format_sensor(sensor, as_json=False)

        # Snapshot the text output
        snapshot.assert_match(result)

    def test_format_sensor_json_output(self, snapshot):
        """Test formatting single sensor as JSON."""
        sensor = self._create_sample_sensor()
        result = format_sensor(sensor, as_json=True)

        # For JSON, we want to snapshot the parsed structure to avoid formatting differences
        parsed = json.loads(result)
        snapshot.assert_match(parsed)

    def test_format_minimal_sensor_text_output(self, snapshot):
        """Test formatting minimal sensor as text."""
        from dagster_shared.utils.timing import fixed_timezone

        sensor = DgApiSensor(
            id="minimal-id",
            name="minimal_sensor",
            status=DgApiSensorStatus.PAUSED,
            sensor_type=DgApiSensorType.STANDARD,
            description=None,
            repository_origin=None,
            next_tick_timestamp=None,
        )
        with fixed_timezone("UTC"):
            result = format_sensor(sensor, as_json=False)

        snapshot.assert_match(result)

    def test_format_minimal_sensor_json_output(self, snapshot):
        """Test formatting minimal sensor as JSON."""
        sensor = DgApiSensor(
            id="minimal-id",
            name="minimal_sensor",
            status=DgApiSensorStatus.PAUSED,
            sensor_type=DgApiSensorType.STANDARD,
            description=None,
            repository_origin=None,
            next_tick_timestamp=None,
        )
        result = format_sensor(sensor, as_json=True)

        parsed = json.loads(result)
        snapshot.assert_match(parsed)


class TestSensorDataProcessing:
    """Test processing of sensor data structures.

    This class tests pure functions and domain model functionality
    without requiring external dependencies.
    """

    def test_sensor_creation_with_all_fields(self, snapshot):
        """Test creating sensor with all possible fields."""
        sensor = DgApiSensor(
            id="complete-sensor-xyz",
            name="comprehensive_sensor",
            status=DgApiSensorStatus.RUNNING,
            sensor_type=DgApiSensorType.FRESHNESS_POLICY,
            description="Comprehensive test sensor with all fields",
            repository_origin="test_location@test_repo",
            next_tick_timestamp=1705311000.0,
        )

        # Test JSON serialization works correctly
        result = sensor.model_dump_json(indent=2)
        parsed = json.loads(result)
        snapshot.assert_match(parsed)

    def test_sensor_status_enum_values(self):
        """Test that all expected SensorStatus enum values are available."""
        expected_statuses = ["RUNNING", "STOPPED", "PAUSED"]

        actual_statuses = [status.value for status in DgApiSensorStatus]
        assert set(actual_statuses) == set(expected_statuses)

    def test_sensor_type_enum_values(self):
        """Test that all expected SensorType enum values are available."""
        expected_types = [
            "STANDARD",
            "MULTI_ASSET",
            "FRESHNESS_POLICY",
            "AUTO_MATERIALIZE",
            "ASSET",
        ]

        actual_types = [sensor_type.value for sensor_type in DgApiSensorType]
        assert set(actual_types) == set(expected_types)

    def test_sensor_with_missing_optional_fields(self):
        """Test sensor creation with None values for optional fields."""
        sensor = DgApiSensor(
            id="sparse-sensor-123",
            name="sparse_sensor",
            status=DgApiSensorStatus.STOPPED,
            sensor_type=DgApiSensorType.STANDARD,
            description=None,
            repository_origin=None,
            next_tick_timestamp=None,
        )

        assert sensor.id == "sparse-sensor-123"
        assert sensor.name == "sparse_sensor"
        assert sensor.status == DgApiSensorStatus.STOPPED
        assert sensor.sensor_type == DgApiSensorType.STANDARD
        assert sensor.description is None
        assert sensor.repository_origin is None
        assert sensor.next_tick_timestamp is None

    def test_sensor_list_creation(self):
        """Test SensorList creation and basic functionality."""
        sensors = [
            DgApiSensor(
                id="sensor1",
                name="first_sensor",
                status=DgApiSensorStatus.RUNNING,
                sensor_type=DgApiSensorType.STANDARD,
            ),
            DgApiSensor(
                id="sensor2",
                name="second_sensor",
                status=DgApiSensorStatus.PAUSED,
                sensor_type=DgApiSensorType.ASSET,
            ),
        ]
        sensor_list = DgApiSensorList(items=sensors, total=len(sensors))

        assert len(sensor_list.items) == 2
        assert sensor_list.total == 2
        assert sensor_list.items[0].name == "first_sensor"
        assert sensor_list.items[1].name == "second_sensor"

    def test_sensor_timestamp_handling(self, snapshot):
        """Test sensor with various timestamp values."""
        test_cases = [
            # Normal timestamp
            DgApiSensor(
                id="sensor1",
                name="normal_timestamp",
                status=DgApiSensorStatus.RUNNING,
                sensor_type=DgApiSensorType.STANDARD,
                next_tick_timestamp=1705311000.0,
            ),
            # No timestamp
            DgApiSensor(
                id="sensor2",
                name="no_timestamp",
                status=DgApiSensorStatus.STOPPED,
                sensor_type=DgApiSensorType.STANDARD,
                next_tick_timestamp=None,
            ),
            # Future timestamp
            DgApiSensor(
                id="sensor3",
                name="future_timestamp",
                status=DgApiSensorStatus.PAUSED,
                sensor_type=DgApiSensorType.STANDARD,
                next_tick_timestamp=2000000000.0,  # Far future
            ),
        ]

        results = []
        for sensor in test_cases:
            # Test serialization
            serialized = json.loads(sensor.model_dump_json())
            results.append(serialized)

        snapshot.assert_match(results)



import textwrap

import click
from click.testing import CliRunner
from dagster_dg_core.utils import DgClickCommand, DgClickGroup, set_option_help_output_group
from dagster_test.dg_utils.utils import (
    ProxyRunner,
    assert_runner_result,
    fixed_panel_width,
    isolated_example_project_foo_bar,
    match_terminal_box_output,
)

# ########################
# ##### TEST CLI
# ########################

# The names of our global options are special-cased, so use one of them here (--verbose) as
# a test case.


@click.group(name="root", cls=DgClickGroup)
@click.option("--root-opt", type=str, default="root", help="Root option.")
@click.option("--verbose", type=str, default="test", help="Verbose output.")
def root(root_opt, verbose):
    """Root group."""
    pass


@root.group(name="sub-group", cls=DgClickGroup)
@click.option("--sub-group-opt", type=str, default="sub-group", help="Sub-group option.")
@click.option("--verbose", type=str, default="test", help="Verbose output.")
def sub_group(sub_group_opt, verbose):
    """Sub-group."""
    pass


@sub_group.command(name="sub-group-command", cls=DgClickCommand)
@click.option(
    "--sub-group-command-opt",
    type=str,
    default="sub_group_command",
    help="Sub-group-command option.",
)
@click.option("--verbose", type=str, default="test", help="Verbose output.")
def sub_group_command(sub_group_command_opt, verbose):
    """Sub-group-command."""
    pass


@root.group(name="sub-command", cls=DgClickGroup)
@click.option("--sub-command-opt", type=str, default="sub-command", help="Sub-command option.")
@click.option("--verbose", type=str, default="test", help="Verbose output.")
def sub_command(sub_command_opt, verbose):
    """Sub-command."""
    pass


for cmd in [root, sub_group, sub_group_command, sub_command]:
    # Make this a global option
    verbose_opt = next(p for p in cmd.params if p.name == "verbose")
    set_option_help_output_group(verbose_opt, "Global options")


# ########################
# ##### TESTS
# ########################


def test_root_help_message():
    runner = CliRunner()
    with fixed_panel_width():
        result = runner.invoke(root, ["--help"])
    assert_runner_result(result)

    assert match_terminal_box_output(
        result.output.strip(),
        textwrap.dedent("""
             Usage: root [OPTIONS] COMMAND [ARGS]...

             Root group.

            ╭─ Options ────────────────────────────────────────────────────────────────────╮
            │ --root-opt        TEXT  Root option.                                         │
            │ --help                  Show this message and exit.                          │
            ╰──────────────────────────────────────────────────────────────────────────────╯
            ╭─ Global options ─────────────────────────────────────────────────────────────╮
            │ --verbose        TEXT  Verbose output.                                       │
            ╰──────────────────────────────────────────────────────────────────────────────╯
            ╭─ Commands ───────────────────────────────────────────────────────────────────╮
            │ sub-command  Sub-command.                                                    │
            │ sub-group    Sub-group.                                                      │
            ╰──────────────────────────────────────────────────────────────────────────────╯
    """).strip(),
    )


def test_sub_group_with_option_help_message():
    runner = CliRunner()
    with fixed_panel_width():
        result = runner.invoke(root, ["sub-group", "--help"])
    assert_runner_result(result)
    assert match_terminal_box_output(
        result.output.strip(),
        textwrap.dedent("""
             Usage: root sub-group [OPTIONS] COMMAND [ARGS]...

             Sub-group.

            ╭─ Options ────────────────────────────────────────────────────────────────────╮
            │ --sub-group-opt        TEXT  Sub-group option.                               │
            │ --help                       Show this message and exit.                     │
            ╰──────────────────────────────────────────────────────────────────────────────╯
            ╭─ Global options ─────────────────────────────────────────────────────────────╮
            │ --verbose        TEXT  Verbose output.                                       │
            ╰──────────────────────────────────────────────────────────────────────────────╯
            ╭─ Commands ───────────────────────────────────────────────────────────────────╮
            │ sub-group-command  Sub-group-command.                                        │
            ╰──────────────────────────────────────────────────────────────────────────────╯
        """).strip(),
    )


def test_sub_group_command_with_option_help_message():
    runner = CliRunner()
    with fixed_panel_width():
        result = runner.invoke(root, ["sub-group", "sub-group-command", "--help"])
    assert_runner_result(result)
    assert match_terminal_box_output(
        result.output.strip(),
        textwrap.dedent("""
             Usage: root sub-group sub-group-command [OPTIONS]

             Sub-group-command.

            ╭─ Options ────────────────────────────────────────────────────────────────────╮
            │ --sub-group-command-opt        TEXT  Sub-group-command option.               │
            │ --help                               Show this message and exit.             │
            ╰──────────────────────────────────────────────────────────────────────────────╯
            ╭─ Global options ─────────────────────────────────────────────────────────────╮
            │ --verbose        TEXT  Verbose output.                                       │
            ╰──────────────────────────────────────────────────────────────────────────────╯
    """).strip(),
    )


def test_sub_command_with_option_help_message():
    runner = CliRunner()
    with fixed_panel_width():
        result = runner.invoke(root, ["sub-command", "--help"])
    assert_runner_result(result)
    assert match_terminal_box_output(
        result.output.strip(),
        textwrap.dedent("""
             Usage: root sub-command [OPTIONS] COMMAND [ARGS]...

             Sub-command.

            ╭─ Options ────────────────────────────────────────────────────────────────────╮
            │ --sub-command-opt        TEXT  Sub-command option.                           │
            │ --help                         Show this message and exit.                   │
            ╰──────────────────────────────────────────────────────────────────────────────╯
            ╭─ Global options ─────────────────────────────────────────────────────────────╮
            │ --verbose        TEXT  Verbose output.                                       │
            ╰──────────────────────────────────────────────────────────────────────────────╯
    """).strip(),
    )


def test_dynamic_subcommand_help_message():
    with (
        ProxyRunner.test(use_fixed_test_components=True) as runner,
        isolated_example_project_foo_bar(runner, in_workspace=False),
    ):
        with fixed_panel_width(width=120):
            result = runner.invoke(
                "scaffold", "defs", "dagster_test.components.SimplePipesScriptComponent", "--help"
            )
            assert_runner_result(result)
            assert match_terminal_box_output(
                result.output.strip(),
                textwrap.dedent("""
                Usage: dg scaffold defs [GLOBAL OPTIONS] dagster_test.components.SimplePipesScriptComponent [OPTIONS] DEFS_PATH

                A simple asset that runs a Python script with the Pipes subprocess client.                                             
                                                                                                                                        
                Because it is a pipes asset, no value is returned.

                ╭─ Arguments ──────────────────────────────────────────────────────────────────────────────────────────────────────────╮
                │ *    defs_path      TEXT  [required]                                                                                 │
                ╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯
                ╭─ Options ────────────────────────────────────────────────────────────────────────────────────────────────────────────╮
                │ --format               [yaml|python]  Format of the component configuration (yaml or python)                         │
                │ --json-params          TEXT           JSON string of scaffolder parameters. Mutually exclusive with passing          │
                │                                       individual parameters as options.                                              │
                │ --asset-key            TEXT           (scaffolder param) asset_key                                                   │
                │ --filename             TEXT           (scaffolder param) filename                                                    │
                │ --help         -h                     Show this message and exit.                                                    │
                ╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯
                ╭─ Global options ─────────────────────────────────────────────────────────────────────────────────────────────────────╮
                │ --verbose          Enable verbose output for debugging.                                                              │
                ╰──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯
                """).strip(),
            )



"""Run events API implementation."""

from dataclasses import dataclass
from typing import TYPE_CHECKING

from dagster_dg_cli.api_layer.graphql_adapter.run_event import get_run_events_via_graphql
from dagster_dg_cli.utils.plus.gql_client import IGraphQLClient

if TYPE_CHECKING:
    from dagster_dg_cli.api_layer.schemas.run_event import RunEventList


@dataclass(frozen=True)
class DgApiRunEventApi:
    """API for run events operations."""

    client: IGraphQLClient

    def get_events(
        self,
        run_id: str,
        event_types: tuple[str, ...] = (),
        step_keys: tuple[str, ...] = (),
        levels: tuple[str, ...] = (),
        limit: int = 100,
        after_cursor: str | None = None,
    ) -> "RunEventList":
        """Get run events with filtering options."""
        from dagster_dg_cli.api_layer.schemas.run_event import (
            DgApiErrorInfo,
            DgApiRunEvent,
            RunEventLevel,
            RunEventList,
        )

        events_data = get_run_events_via_graphql(
            self.client,
            run_id=run_id,
            limit=limit,
            after_cursor=after_cursor,
            event_types=event_types,
            step_keys=step_keys,
            levels=levels,
        )

        # Helper function to convert error data to DgApiErrorInfo recursively
        def _convert_error_info(error_data: dict | None) -> DgApiErrorInfo | None:
            if not error_data:
                return None
            return DgApiErrorInfo(
                message=error_data.get("message", ""),
                className=error_data.get("className"),
                stack=error_data.get("stack"),
                cause=_convert_error_info(error_data.get("cause")),
            )

        # Convert to Pydantic models
        events = [
            DgApiRunEvent(
                run_id=e["runId"],
                message=e["message"],
                timestamp=e["timestamp"],
                level=RunEventLevel[e["level"]],
                step_key=e.get("stepKey"),
                event_type=e.get("eventType"),
                error=_convert_error_info(e.get("error")),
            )
            for e in events_data["events"]
        ]

        return RunEventList(
            items=events,
            total=len(events),
            cursor=events_data.get("cursor"),
            has_more=events_data.get("hasMore", False),
        )



"""Tests for the diagnostics module."""

import json
import tempfile
from pathlib import Path

from dagster_dg_cli.cli.scaffold.branch.claude.diagnostics import (
    AIInteraction,
    ClaudeDiagnostics,
    ContextGathering,
    DiagnosticsEntry,
    PerformanceMetrics,
    create_claude_diagnostics_service,
)


class TestClaudeDiagnosticsService:
    def test_initialization_off_by_default(self):
        service = ClaudeDiagnostics()
        assert service.level == "off"
        assert service.correlation_id is not None
        assert service.entries == []

    def test_initialization_with_custom_params(self):
        correlation_id = "test-correlation-id"
        output_dir = Path("/tmp/test")
        service = ClaudeDiagnostics(
            level="debug",
            output_dir=output_dir,
            correlation_id=correlation_id,
        )
        assert service.level == "debug"
        assert service.correlation_id == correlation_id
        assert service.output_dir == output_dir

    def test_should_log_hierarchy(self):
        # Test logging hierarchy: error < info < debug
        service = ClaudeDiagnostics(level="info")

        assert service._should_log("error") is True  # noqa: SLF001
        assert service._should_log("info") is True  # noqa: SLF001
        assert service._should_log("debug") is False  # noqa: SLF001

        service.level = "debug"
        assert service._should_log("error") is True  # noqa: SLF001
        assert service._should_log("info") is True  # noqa: SLF001
        assert service._should_log("debug") is True  # noqa: SLF001

        service.level = "error"
        assert service._should_log("error") is True  # noqa: SLF001
        assert service._should_log("info") is False  # noqa: SLF001
        assert service._should_log("debug") is False  # noqa: SLF001

    def test_should_not_log_when_off(self):
        service = ClaudeDiagnostics(level="off")

        assert service._should_log("error") is False  # noqa: SLF001
        assert service._should_log("info") is False  # noqa: SLF001
        assert service._should_log("debug") is False  # noqa: SLF001

    def test_log_creates_entry(self):
        service = ClaudeDiagnostics(level="info")

        service.log(
            level="info", category="test_category", message="test message", data={"key": "value"}
        )

        assert len(service.entries) == 1
        entry = service.entries[0]
        assert entry.level == "info"
        assert entry.category == "test_category"
        assert entry.message == "test message"
        assert entry.data == {"key": "value"}
        assert entry.correlation_id == service.correlation_id

    def test_log_does_not_create_entry_when_level_too_low(self):
        service = ClaudeDiagnostics(level="error")

        service.log(level="info", category="test_category", message="test message")

        assert len(service.entries) == 0

    def test_data_logging_without_sanitization(self):
        service = ClaudeDiagnostics(level="info")

        test_data = {
            "api_key": "sk-1234567890abcdef",  # This will no longer be redacted
            "secret_token": "secret-value",
            "password": "mypassword",
            "normal_key": "normal_value",
        }

        service.log(level="info", category="test", message="message", data=test_data)

        entry = service.entries[0]
        # Data should be logged as-is without sanitization
        assert entry.data["api_key"] == "sk-1234567890abcdef"
        assert entry.data["secret_token"] == "secret-value"
        assert entry.data["password"] == "mypassword"
        assert entry.data["normal_key"] == "normal_value"

    def test_ai_interaction_logging(self):
        service = ClaudeDiagnostics(level="info")

        interaction = AIInteraction(
            correlation_id="test-id",
            timestamp="2024-01-01T00:00:00",
            prompt="test prompt",
            response="test response",
            token_count=100,
            allowed_tools=["tool1", "tool2"],
            duration_ms=500.0,
        )

        service.log_ai_interaction(interaction)

        assert len(service.entries) == 1
        entry = service.entries[0]
        assert entry.category == "ai_interaction"
        assert entry.data["prompt_length"] == len("test prompt")
        assert entry.data["response_length"] == len("test response")
        assert entry.data["token_count"] == 100
        assert entry.data["allowed_tools"] == ["tool1", "tool2"]
        assert entry.data["duration_ms"] == 500.0

    def test_context_gathering_logging(self):
        service = ClaudeDiagnostics(level="debug")

        context = ContextGathering(
            correlation_id="test-id",
            timestamp="2024-01-01T00:00:00",
            files_analyzed=["file1.py", "file2.py"],
            patterns_detected=["pattern1", "pattern2"],
            decisions_made={"decision1": "value1"},
        )

        service.log_context_gathering(context)

        assert len(service.entries) == 1
        entry = service.entries[0]
        assert entry.category == "context_gathering"
        assert entry.data["files_count"] == 2
        assert entry.data["files_analyzed"] == ["file1.py", "file2.py"]
        assert entry.data["patterns_detected"] == ["pattern1", "pattern2"]

    def test_performance_logging(self):
        service = ClaudeDiagnostics(level="debug")

        metrics = PerformanceMetrics(
            correlation_id="test-id",
            timestamp="2024-01-01T00:00:00",
            operation="test_operation",
            duration_ms=250.0,
            phase="test_phase",
        )

        service.log_performance(metrics)

        assert len(service.entries) == 1
        entry = service.entries[0]
        assert entry.category == "performance"
        assert entry.data["operation"] == "test_operation"
        assert entry.data["phase"] == "test_phase"
        assert entry.data["duration_ms"] == 250.0

    def test_time_operation_context_manager(self):
        service = ClaudeDiagnostics(level="debug")

        with service.time_operation("test_op", "test_phase"):
            # Simulate some work
            pass

        assert len(service.entries) == 1
        entry = service.entries[0]
        assert entry.category == "performance"
        assert entry.data["operation"] == "test_op"
        assert entry.data["phase"] == "test_phase"
        assert entry.data["duration_ms"] >= 0

    def test_convenience_methods(self):
        service = ClaudeDiagnostics(level="debug")

        service.error(category="test_cat", message="error message")
        service.info(category="test_cat", message="info message")
        service.debug(category="test_cat", message="debug message")

        assert len(service.entries) == 3
        assert service.entries[0].level == "error"
        assert service.entries[1].level == "info"
        assert service.entries[2].level == "debug"

    def test_flush_writes_file(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            output_dir = Path(temp_dir)
            service = ClaudeDiagnostics(level="info", output_dir=output_dir)

            service.info(category="test", message="test message", data={"key": "value"})

            output_path = service.flush()

            assert output_path is not None
            assert output_path.exists()

            # Verify file contents - parse JSONL format
            lines = output_path.read_text().strip().split("\n")
            assert len(lines) >= 3  # session_start, entry, session_end

            # Parse each line as JSON
            session_start = json.loads(lines[0])
            entry = json.loads(lines[1])
            session_end = json.loads(lines[2])

            # Verify session_start
            assert session_start["type"] == "session_start"
            assert session_start["correlation_id"] == service.correlation_id
            assert session_start["level"] == "info"

            # Verify entry
            assert entry["type"] == "entry"
            assert entry["correlation_id"] == service.correlation_id
            assert entry["level"] == "info"
            assert entry["category"] == "test"
            assert entry["message"] == "test message"
            assert entry["data"] == {"key": "value"}

            # Verify session_end
            assert session_end["type"] == "session_end"
            assert session_end["correlation_id"] == service.correlation_id

            # Verify entries are cleared after flush
            assert len(service.entries) == 0

    def test_flush_returns_none_when_off(self):
        service = ClaudeDiagnostics(level="off")
        service.entries = []  # Ensure no entries

        output_path = service.flush()

        assert output_path is None

    def test_flush_returns_none_when_no_entries(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            output_dir = Path(temp_dir)
            service = ClaudeDiagnostics(level="info", output_dir=output_dir)

            output_path = service.flush()

            assert output_path is None


class TestClaudeDiagnosticsServiceCreation:
    def test_create_diagnostics_service_defaults(self):
        service = create_claude_diagnostics_service()

        assert service.level == "off"
        assert service.correlation_id is not None
        assert service.output_dir == Path(tempfile.gettempdir()) / "dg" / "diagnostics"

    def test_create_diagnostics_service_with_params(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            output_dir = Path(temp_dir)
            correlation_id = "test-correlation-id"

            service = create_claude_diagnostics_service(
                level="debug",
                output_dir=output_dir,
                correlation_id=correlation_id,
            )

            assert service.level == "debug"
            assert service.correlation_id == correlation_id
            assert service.output_dir == output_dir

    def test_create_diagnostics_service_string_path(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            service = create_claude_diagnostics_service(
                level="info",
                output_dir=temp_dir,
            )

            assert service.level == "info"
            assert service.output_dir == Path(temp_dir)


class TestDiagnosticsDataModels:
    def test_diagnostics_entry_creation(self):
        entry = DiagnosticsEntry(
            correlation_id="test-id",
            timestamp="2024-01-01T00:00:00",
            level="info",
            category="test",
            message="test message",
            data={"key": "value"},
        )

        assert entry.correlation_id == "test-id"
        assert entry.timestamp == "2024-01-01T00:00:00"
        assert entry.level == "info"
        assert entry.category == "test"
        assert entry.message == "test message"
        assert entry.data == {"key": "value"}

    def test_ai_interaction_creation(self):
        interaction = AIInteraction(
            correlation_id="test-id",
            timestamp="2024-01-01T00:00:00",
            prompt="test prompt",
            response="test response",
            token_count=100,
            allowed_tools=["tool1"],
            duration_ms=500.0,
        )

        assert interaction.prompt == "test prompt"
        assert interaction.response == "test response"
        assert interaction.token_count == 100
        assert interaction.allowed_tools == ["tool1"]
        assert interaction.duration_ms == 500.0

    def test_context_gathering_creation(self):
        context = ContextGathering(
            correlation_id="test-id",
            timestamp="2024-01-01T00:00:00",
            files_analyzed=["file1.py"],
            patterns_detected=["pattern1"],
            decisions_made={"key": "value"},
        )

        assert context.files_analyzed == ["file1.py"]
        assert context.patterns_detected == ["pattern1"]
        assert context.decisions_made == {"key": "value"}

    def test_performance_metrics_creation(self):
        metrics = PerformanceMetrics(
            correlation_id="test-id",
            timestamp="2024-01-01T00:00:00",
            operation="test_op",
            duration_ms=250.0,
            phase="test_phase",
        )

        assert metrics.operation == "test_op"
        assert metrics.duration_ms == 250.0
        assert metrics.phase == "test_phase"

    def test_claude_operation_success_logging(self):
        """Test that the error boundary logs start and success messages automatically."""
        service = ClaudeDiagnostics(level="info")

        with service.claude_operation(
            operation_name="test_operation",
            error_code="test_error",
            error_message="Test failed",
        ):
            # Simulate some work
            pass

        # Should have 2 entries: start and success
        assert len(service.entries) == 2

        start_entry = service.entries[0]
        assert start_entry.level == "info"
        assert start_entry.category == "test_operation_start"
        assert start_entry.message == "Starting test_operation"

        success_entry = service.entries[1]
        assert success_entry.level == "info"
        assert success_entry.category == "test_operation_success"
        assert success_entry.message == "Successfully completed test_operation"
        assert "duration_ms" in success_entry.data

    def test_claude_operation_error_logging(self):
        """Test that the error boundary logs errors when exceptions occur."""
        service = ClaudeDiagnostics(level="info")

        try:
            with service.claude_operation(
                operation_name="failing_operation",
                error_code="operation_failed",
                error_message="Operation failed with error",
            ):
                raise ValueError("Test error")
        except ValueError:
            pass  # Expected

        # Should have 2 entries: start and error (no success)
        assert len(service.entries) == 2

        start_entry = service.entries[0]
        assert start_entry.level == "info"
        assert start_entry.category == "failing_operation_start"
        assert start_entry.message == "Starting failing_operation"

        error_entry = service.entries[1]
        assert error_entry.level == "error"
        assert error_entry.category == "operation_failed"
        assert error_entry.message == "Operation failed with error"
        assert error_entry.data["error_type"] == "ValueError"
        assert error_entry.data["error_message"] == "Test error"
        assert "duration_ms" in error_entry.data



"""Asset API commands following GitHub CLI patterns."""

import json
from typing import Final

import click
from dagster_dg_core.utils import DgClickCommand, DgClickGroup
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper
from dagster_shared.plus.config import DagsterPlusCliConfig
from dagster_shared.plus.config_utils import dg_api_options

from dagster_dg_cli.cli.api.client import create_dg_api_graphql_client
from dagster_dg_cli.cli.api.formatters import format_asset, format_assets

DG_API_MAX_ASSET_LIMIT: Final = 1000


@click.command(name="list", cls=DgClickCommand)
@click.option(
    "--limit",
    type=click.IntRange(1, DG_API_MAX_ASSET_LIMIT),
    default=50,
    help=f"Number of assets to return (default: 50, max: {DG_API_MAX_ASSET_LIMIT})",
)
@click.option(
    "--cursor",
    type=str,
    help="Cursor for pagination",
)
@click.option(
    "--view",
    type=click.Choice(["status"]),
    help="View type: 'status' for health and runtime information",
)
@click.option(
    "--json",
    "output_json",
    is_flag=True,
    help="Output in JSON format for machine readability",
)
@dg_api_options(deployment_scoped=True)
@cli_telemetry_wrapper
@click.pass_context
def list_assets_command(
    ctx: click.Context,
    limit: int,
    cursor: str,
    view: str,
    output_json: bool,
    organization: str,
    deployment: str,
    api_token: str,
    view_graphql: bool,
) -> None:
    """List assets with pagination."""
    config = DagsterPlusCliConfig.create_for_deployment(
        deployment=deployment,
        organization=organization,
        user_token=api_token,
    )
    client = create_dg_api_graphql_client(ctx, config, view_graphql=view_graphql)
    from dagster_dg_cli.api_layer.api.asset import DgApiAssetApi

    api = DgApiAssetApi(client)

    try:
        assets = api.list_assets(limit=limit, cursor=cursor, view=view)
        output = format_assets(assets, as_json=output_json)
        click.echo(output)
    except Exception as e:
        if output_json:
            error_response = {"error": str(e)}
            click.echo(json.dumps(error_response), err=True)
        else:
            click.echo(f"Error querying Dagster Plus API: {e}", err=True)
        raise click.ClickException(f"Failed to list assets: {e}")


@click.command(name="get", cls=DgClickCommand)
@click.argument("asset_key", type=str)
@click.option(
    "--view",
    type=click.Choice(["status"]),
    help="View type: 'status' for health and runtime information",
)
@click.option(
    "--json",
    "output_json",
    is_flag=True,
    help="Output in JSON format for machine readability",
)
@dg_api_options(deployment_scoped=True)
@cli_telemetry_wrapper
@click.pass_context
def get_asset_command(
    ctx: click.Context,
    asset_key: str,
    view: str,
    output_json: bool,
    organization: str,
    deployment: str,
    api_token: str,
    view_graphql: bool,
) -> None:
    """Get specific asset details."""
    config = DagsterPlusCliConfig.create_for_deployment(
        deployment=deployment,
        organization=organization,
        user_token=api_token,
    )
    client = create_dg_api_graphql_client(ctx, config, view_graphql=view_graphql)
    from dagster_dg_cli.api_layer.api.asset import DgApiAssetApi

    api = DgApiAssetApi(client)

    try:
        asset = api.get_asset(asset_key, view=view)
        output = format_asset(asset, as_json=output_json)
        click.echo(output)
    except Exception as e:
        if output_json:
            error_response = {"error": str(e)}
            click.echo(json.dumps(error_response), err=True)
        else:
            click.echo(f"Error querying Dagster Plus API: {e}", err=True)
        raise click.ClickException(f"Failed to get asset: {e}")


@click.group(
    name="asset",
    cls=DgClickGroup,
    commands={
        "list": list_assets_command,
        "get": get_asset_command,
    },
)
def asset_group():
    """Manage assets in Dagster Plus."""



"""Implementation for configuring build artifacts (Dockerfiles, build configs, etc)."""

import click

from dagster_dg_cli.cli.plus.constants import DgPlusAgentType
from dagster_dg_cli.cli.plus.deploy.configure.utils import (
    DgPlusDeployConfigureOptions,
    display_registry_secrets_hints,
    get_project_contexts,
    get_scaffolded_container_context_yaml,
    prompt_for_registry_url,
)
from dagster_dg_cli.utils.plus.build import create_deploy_dockerfile, get_dockerfile_path

# Registry comment block with examples
REGISTRY_COMMENTS = """\
# Container registry URL for pushing Docker images.
# Examples:
#   ECR:       123456789012.dkr.ecr.us-east-1.amazonaws.com/my-repo
#   GCR:       gcr.io/my-project/my-image
#   Azure:     myregistry.azurecr.io/my-image
#   DockerHub: docker.io/myuser/my-image
#   GHCR:      ghcr.io/myorg/my-image
"""

PROJECT_BUILD_YAML_IN_WORKSPACE = """\
directory: .
# Registry is specified in the build.yaml file at the root of the workspace,
# but can be overridden here.
# registry: '...'
"""


def configure_build_artifacts_impl(config: DgPlusDeployConfigureOptions) -> None:
    """Core implementation for configuring build artifacts."""
    # For serverless with PEX deploys, skip scaffolding build artifacts
    if config.agent_type == DgPlusAgentType.SERVERLESS and config.pex_deploy:
        return

    scaffolded_container_context_yaml = (
        get_scaffolded_container_context_yaml(config.agent_platform)
        if config.agent_platform
        else None
    )

    # For hybrid deployments, use provided registry URL or prompt for one
    registry_url: str | None = None
    if config.agent_type == DgPlusAgentType.HYBRID:
        if config.registry_url:
            registry_url = config.registry_url
        else:
            registry_url = prompt_for_registry_url(skip_prompt=config.skip_confirmation_prompt)

    # Use the provided registry URL or fall back to placeholder
    registry_value = registry_url if registry_url else "..."

    if not config.dg_context.is_project:
        click.echo("Scaffolding build artifacts for workspace...")

        create = True
        if config.dg_context.build_config_path.exists():
            create = config.skip_confirmation_prompt or click.confirm(
                f"Build config already exists at {config.dg_context.build_config_path}. Overwrite it?",
            )
        if create:
            build_yaml_content = f"{REGISTRY_COMMENTS}registry: '{registry_value}'\n"
            config.dg_context.build_config_path.write_text(build_yaml_content)
            click.echo(f"Workspace build config created at {config.dg_context.build_config_path}.")
            if registry_url:
                click.echo(f"Registry configured: {registry_url}")

        if scaffolded_container_context_yaml:
            create = True
            if config.dg_context.container_context_config_path.exists():
                create = config.skip_confirmation_prompt or click.confirm(
                    f"Container config already exists at {config.dg_context.container_context_config_path}. Overwrite it?",
                )
            if create:
                config.dg_context.container_context_config_path.write_text(
                    scaffolded_container_context_yaml
                )
                click.echo(
                    f"Workspace container config created at {config.dg_context.container_context_config_path}."
                )

    project_contexts = get_project_contexts(config.dg_context, config.cli_config)
    for project_context in project_contexts:
        click.echo(f"Scaffolding build artifacts for {project_context.code_location_name}...")

        create = True
        if project_context.build_config_path.exists():
            create = config.skip_confirmation_prompt or click.confirm(
                f"Build config already exists at {project_context.build_config_path}. Overwrite it?",
            )
        if create:
            if config.dg_context.is_in_workspace:
                # In a workspace, registry is at workspace level
                project_context.build_config_path.write_text(PROJECT_BUILD_YAML_IN_WORKSPACE)
            else:
                # Standalone project needs its own registry
                project_build_yaml = (
                    f"directory: .\n{REGISTRY_COMMENTS}registry: '{registry_value}'\n"
                )
                project_context.build_config_path.write_text(project_build_yaml)
                if registry_url:
                    click.echo(f"Registry configured: {registry_url}")
            click.echo(f"Project build config created at {project_context.build_config_path}.")

        if scaffolded_container_context_yaml:
            if project_context.container_context_config_path.exists():
                create = config.skip_confirmation_prompt or click.confirm(
                    f"Container config already exists at {project_context.container_context_config_path}. Overwrite it?",
                )
            if create:
                project_context.container_context_config_path.write_text(
                    scaffolded_container_context_yaml
                )
                click.echo(
                    f"Project container config created at {project_context.container_context_config_path}."
                )

        dockerfile_path = get_dockerfile_path(project_context, workspace_context=config.dg_context)
        create = True
        if dockerfile_path.exists():
            create = config.skip_confirmation_prompt or click.confirm(
                f"A Dockerfile already exists at {dockerfile_path}. Overwrite it?",
            )
        if create:
            # python_version should always be resolved, but check for safety
            if config.python_version is None:
                raise click.ClickException(
                    "python_version is required for creating Dockerfiles. "
                    "Please specify --python-version."
                )
            create_deploy_dockerfile(
                dockerfile_path,
                config.python_version,
                config.use_editable_dagster,
                project_context.package_name,
            )
            click.echo(f"Dockerfile created at {dockerfile_path}.")

    # Display secrets hints if registry was configured
    if config.agent_type == DgPlusAgentType.HYBRID and registry_url:
        display_registry_secrets_hints(registry_url, config.git_provider)



"""Sensor API implementation."""

from dataclasses import dataclass
from typing import TYPE_CHECKING

from dagster_dg_cli.api_layer.graphql_adapter.sensor import (
    get_sensor_via_graphql,
    list_sensors_via_graphql,
)
from dagster_dg_cli.utils.plus.gql_client import IGraphQLClient

if TYPE_CHECKING:
    from dagster_dg_cli.api_layer.schemas.sensor import DgApiSensor, DgApiSensorList


@dataclass(frozen=True)
class DgApiSensorApi:
    """API for sensor operations."""

    client: IGraphQLClient

    def list_sensors(
        self,
        repository_location_name: str | None = None,
        repository_name: str | None = None,
    ) -> "DgApiSensorList":
        """List all sensors, optionally filtered by repository location and name."""
        return list_sensors_via_graphql(
            self.client,
            repository_location_name=repository_location_name,
            repository_name=repository_name,
        )

    def get_sensor(
        self,
        sensor_name: str,
        repository_location_name: str,
        repository_name: str,
    ) -> "DgApiSensor":
        """Get sensor by name and repository details."""
        return get_sensor_via_graphql(
            self.client,
            sensor_name=sensor_name,
            repository_location_name=repository_location_name,
            repository_name=repository_name,
        )

    def get_sensor_by_name(self, sensor_name: str) -> "DgApiSensor":
        """Get sensor by name, searching across all repositories."""
        from dagster_dg_cli.api_layer.graphql_adapter.sensor import get_sensor_by_name_via_graphql

        return get_sensor_by_name_via_graphql(self.client, sensor_name=sensor_name)



"""Primary interface for deployment configuration scaffolding.

This module provides the `dg scaffold deployment-config` command group with
subcommands for serverless and hybrid deployments.
"""

import sys
from pathlib import Path

import click
from dagster_cloud_cli.utils import SUPPORTED_PYTHON_VERSIONS
from dagster_dg_core.config import normalize_cli_config
from dagster_dg_core.context import DgContext
from dagster_dg_core.shared_options import dg_editable_dagster_options, dg_global_options
from dagster_dg_core.utils import DgClickCommand, DgClickGroup, exit_with_error
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper
from dagster_shared.plus.config import DAGSTER_CLOUD_BASE_URL, DagsterPlusCliConfig

from dagster_dg_cli.cli.plus.constants import DgPlusAgentPlatform, DgPlusAgentType
from dagster_dg_cli.cli.plus.deploy.configure.configure_build_artifacts import (
    configure_build_artifacts_impl,
)
from dagster_dg_cli.cli.plus.deploy.configure.configure_ci import configure_ci_impl
from dagster_dg_cli.cli.plus.deploy.configure.utils import (
    DgPlusDeployConfigureOptions,
    GitProvider,
    detect_agent_type_and_platform,
    search_for_git_root,
)


def resolve_agent_type_and_platform(
    agent_type: DgPlusAgentType | None,
    agent_platform: DgPlusAgentPlatform | None,
    plus_config: DagsterPlusCliConfig | None,
) -> tuple[DgPlusAgentType, DgPlusAgentPlatform | None]:
    """Resolve agent type and platform from config or prompts (for deployment-config commands)."""
    resolved_type = agent_type
    resolved_platform = agent_platform

    # Try to detect from Plus config via GraphQL
    if resolved_type is None and plus_config:
        detected_type, detected_platform = detect_agent_type_and_platform(plus_config)
        resolved_type = detected_type
        if resolved_platform is None:
            resolved_platform = detected_platform

    # Prompt for agent type if still missing
    if resolved_type is None:
        resolved_type = DgPlusAgentType(
            click.prompt(
                "Deployment agent type",
                type=click.Choice(["serverless", "hybrid"]),
            ).upper()
        )

    # Prompt for platform if needed (only for hybrid)
    if resolved_type == DgPlusAgentType.HYBRID and resolved_platform is None:
        resolved_platform = DgPlusAgentPlatform(
            click.prompt(
                "Agent platform",
                type=click.Choice(["k8s", "ecs", "docker"]),
            ).upper()
        )

    return resolved_type, resolved_platform


def resolve_organization(
    organization: str | None,
    plus_config: DagsterPlusCliConfig | None,
    *,
    show_detected_message: bool = True,
) -> str:
    """Resolve organization name from config or prompt."""
    if organization is not None:
        return organization

    if plus_config and plus_config.organization:
        if show_detected_message:
            click.echo(
                f"Using organization name {plus_config.organization} from Dagster Plus config."
            )
        return plus_config.organization

    return click.prompt("Dagster Plus organization name") or ""


def resolve_url(
    url: str | None,
    plus_config: DagsterPlusCliConfig | None,
    resolved_organization: str,
    *,
    show_detected_message: bool = True,
) -> str:
    """Resolve Dagster Cloud URL from config or use default."""
    if url is not None:
        return url

    if plus_config and plus_config.url:
        if show_detected_message:
            click.echo(f"Using Dagster Cloud URL {plus_config.url} from Dagster Plus config.")
        return plus_config.url

    return f"{DAGSTER_CLOUD_BASE_URL}/{resolved_organization}"


def resolve_deployment(
    deployment: str | None,
    plus_config: DagsterPlusCliConfig | None,
    *,
    show_detected_message: bool = True,
) -> str:
    """Resolve deployment name from config or prompt."""
    if deployment is not None:
        return deployment

    if plus_config and plus_config.default_deployment:
        if show_detected_message:
            click.echo(
                f"Using default deployment name {plus_config.default_deployment} from Dagster Plus config."
            )
        return plus_config.default_deployment

    return click.prompt("Default deployment name", default="prod")


def resolve_git_provider(
    git_provider: GitProvider | None,
    git_root: Path | None,
) -> GitProvider | None:
    """Resolve git provider for CI/CD scaffolding (for deployment-config commands)."""
    if git_provider is not None:
        return git_provider

    # Prompt for CI/CD configuration
    if git_root is None:
        should_scaffold_ci = click.confirm(
            "Would you like to scaffold CI/CD configuration?", default=True
        )
        if should_scaffold_ci:
            provider_choice = click.prompt(
                "Git provider",
                type=click.Choice(["github", "gitlab"]),
                default="github",
            )
            return GitProvider(provider_choice)

    return None


def resolve_git_root(
    git_root: Path | None,
    git_provider: GitProvider | None,
) -> Path | None:
    """Resolve git root path."""
    if git_provider is None:
        return None

    if git_root is not None:
        return git_root

    resolved_git_root = search_for_git_root(Path.cwd())
    if resolved_git_root is None:
        exit_with_error(
            "No git repository found. Must be run from a git repository, or "
            "specify the path to the git root with `--git-root`."
        )

    return resolved_git_root


def resolve_python_version(python_version: str | None) -> str:
    """Resolve Python version."""
    return python_version or f"3.{sys.version_info.minor}"


def _resolve_config_with_prompts(
    agent_type: DgPlusAgentType | None,
    agent_platform: DgPlusAgentPlatform | None,
    organization: str | None,
    url: str | None,
    deployment: str | None,
    git_root: Path | None,
    python_version: str | None,
    skip_confirmation_prompt: bool,
    use_editable_dagster: bool,
    git_provider: GitProvider | None,
    dg_context: DgContext,
    cli_config,
    pex_deploy: bool | None = None,
    registry_url: str | None = None,
) -> DgPlusDeployConfigureOptions:
    """Resolve all configuration for deployment-config commands, prompting for missing values.

    This is used by the primary deployment-config commands (serverless/hybrid).
    For legacy commands, use the specialized resolve functions below.
    """
    plus_config = DagsterPlusCliConfig.get() if DagsterPlusCliConfig.exists() else None

    # Resolve agent type and platform
    resolved_agent_type, resolved_agent_platform = resolve_agent_type_and_platform(
        agent_type,
        agent_platform,
        plus_config,
    )

    # Resolve git provider
    resolved_git_provider = resolve_git_provider(
        git_provider,
        git_root,
    )

    # Resolve organization, URL, and deployment (only needed if scaffolding CI/CD)
    resolved_organization = None
    resolved_url = None
    resolved_deployment = deployment or "prod"
    if resolved_git_provider is not None:
        resolved_organization = resolve_organization(organization, plus_config)
        resolved_url = resolve_url(url, plus_config, resolved_organization)
        resolved_deployment = resolve_deployment(deployment, plus_config)

    # Resolve git root
    resolved_git_root = resolve_git_root(git_root, resolved_git_provider)

    # Resolve Python version (required for both serverless and hybrid Dockerfile creation)
    resolved_python_version = resolve_python_version(python_version)

    # Resolve pex_deploy for serverless (prompt if not provided)
    resolved_pex_deploy = None
    if resolved_agent_type == DgPlusAgentType.SERVERLESS:
        if pex_deploy is None:
            resolved_pex_deploy = click.confirm(
                "Enable PEX-based fast deploys?",
                default=True,
            )
        else:
            resolved_pex_deploy = pex_deploy

    return DgPlusDeployConfigureOptions(
        dg_context=dg_context,
        cli_config=cli_config,
        plus_config=plus_config,
        agent_type=resolved_agent_type,
        agent_platform=resolved_agent_platform,
        organization_name=resolved_organization,
        cloud_url=resolved_url,
        deployment_name=resolved_deployment,
        git_root=resolved_git_root,
        python_version=resolved_python_version,
        skip_confirmation_prompt=skip_confirmation_prompt,
        git_provider=resolved_git_provider,
        use_editable_dagster=use_editable_dagster,
        pex_deploy=resolved_pex_deploy,
        registry_url=registry_url,
    )


# ########################
# ##### CLI COMMANDS
# ########################


@click.group(name="configure", cls=DgClickGroup, invoke_without_command=True)
@click.option(
    "--git-provider",
    type=click.Choice(["github", "gitlab"]),
    help="Git provider for CI/CD scaffolding",
)
@dg_global_options
@cli_telemetry_wrapper
@click.pass_context
def deploy_configure_group(
    ctx: click.Context,
    git_provider: str | None,
    **global_options: object,
) -> None:
    """Scaffold deployment configuration files for Dagster Plus.

    If no subcommand is specified, will attempt to auto-detect the agent type from your
    Dagster Plus deployment. If detection fails, you will be prompted to choose between
    serverless or hybrid.

    If run in a `workspace <https://docs.dagster.io/guides/build/projects/workspaces/creating-workspaces>`__,
    will scaffold configuration files for the workspace and all projects contained in it.
    """
    # If a subcommand was invoked, let Click handle it
    if ctx.invoked_subcommand is not None:
        return

    cli_config = normalize_cli_config(global_options, ctx)
    dg_context = DgContext.for_workspace_or_project_environment(Path.cwd(), cli_config)

    # Resolve all configuration via prompts and auto-detection
    config = _resolve_config_with_prompts(
        agent_type=None,
        agent_platform=None,
        organization=None,
        url=None,
        deployment=None,
        git_root=None,
        python_version=None,
        skip_confirmation_prompt=False,
        use_editable_dagster=False,
        git_provider=GitProvider(git_provider) if git_provider else None,
        dg_context=dg_context,
        cli_config=cli_config,
        pex_deploy=None,
    )

    configure_build_artifacts_impl(config)
    configure_ci_impl(config)


@click.command(name="serverless", cls=DgClickCommand)
@click.option(
    "--git-provider",
    type=click.Choice(["github", "gitlab"]),
    help="Git provider for CI/CD scaffolding",
)
@click.option(
    "--python-version",
    type=click.Choice(SUPPORTED_PYTHON_VERSIONS),
    help="Python version used to deploy the project",
)
@click.option(
    "--organization",
    help="Dagster Plus organization name",
)
@click.option(
    "--url",
    help="Dagster Plus URL for the organization",
)
@click.option(
    "--deployment",
    default="prod",
    help="Deployment name",
)
@click.option(
    "--git-root",
    type=Path,
    help="Path to the git repository root",
)
@click.option(
    "--pex-deploy/--no-pex-deploy",
    default=True,
    help="Enable PEX-based fast deploys (default: True). If disabled, Docker builds will be used.",
)
@click.option(
    "-y",
    "--yes",
    "skip_confirmation_prompt",
    is_flag=True,
    help="Skip confirmation prompts",
)
@dg_editable_dagster_options
@dg_global_options
@cli_telemetry_wrapper
def deploy_configure_serverless(
    git_provider: str | None,
    python_version: str | None,
    organization: str | None,
    url: str | None,
    deployment: str | None,
    git_root: Path | None,
    pex_deploy: bool,
    skip_confirmation_prompt: bool,
    use_editable_dagster: str | None,
    **global_options: object,
) -> None:
    """Scaffold deployment configuration for Dagster Plus Serverless.

    This creates:
    - Required files for CI/CD based on your Git provider (GitHub Actions or GitLab CI)
    - Dockerfile and build.yaml for containerization (if --no-pex-deploy is used)
    """
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_workspace_or_project_environment(Path.cwd(), cli_config)

    config = _resolve_config_with_prompts(
        agent_type=DgPlusAgentType.SERVERLESS,
        agent_platform=None,
        organization=organization,
        url=url,
        deployment=deployment,
        git_root=git_root,
        python_version=python_version,
        skip_confirmation_prompt=skip_confirmation_prompt,
        use_editable_dagster=bool(use_editable_dagster),
        git_provider=GitProvider(git_provider) if git_provider else None,
        dg_context=dg_context,
        cli_config=cli_config,
        pex_deploy=pex_deploy,
    )

    configure_build_artifacts_impl(config)
    configure_ci_impl(config)


@click.command(name="hybrid", cls=DgClickCommand)
@click.option(
    "--git-provider",
    type=click.Choice(["github", "gitlab"]),
    help="Git provider for CI/CD scaffolding",
)
@click.option(
    "--agent-platform",
    type=click.Choice(["k8s", "ecs", "docker"]),
    help="Agent platform (k8s, ecs, or docker)",
)
@click.option(
    "--registry-url",
    help="Container registry URL for Docker images (e.g., 123456789012.dkr.ecr.us-east-1.amazonaws.com/my-repo)",
)
@click.option(
    "--python-version",
    type=click.Choice(SUPPORTED_PYTHON_VERSIONS),
    help="Python version used to deploy the project",
)
@click.option(
    "--organization",
    help="Dagster Plus organization name",
)
@click.option(
    "--url",
    help="Dagster Cloud URL (defaults to https://dagster.cloud)",
)
@click.option(
    "--deployment",
    default="prod",
    help="Deployment name",
)
@click.option(
    "--git-root",
    type=Path,
    help="Path to the git repository root",
)
@click.option(
    "-y",
    "--yes",
    "skip_confirmation_prompt",
    is_flag=True,
    help="Skip confirmation prompts",
)
@dg_editable_dagster_options
@dg_global_options
@cli_telemetry_wrapper
def deploy_configure_hybrid(
    git_provider: str | None,
    agent_platform: str | None,
    registry_url: str | None,
    python_version: str | None,
    organization: str | None,
    url: str | None,
    deployment: str | None,
    git_root: Path | None,
    skip_confirmation_prompt: bool,
    use_editable_dagster: str | None,
    **global_options: object,
) -> None:
    """Scaffold deployment configuration for Dagster Plus Hybrid.

    This creates:
    - Dockerfile and build.yaml for containerization
    - container_context.yaml with platform-specific config (k8s/ecs/docker)
    - Required files for CI/CD based on your Git provider (GitHub Actions or GitLab CI)
    """
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_workspace_or_project_environment(Path.cwd(), cli_config)

    resolved_platform = DgPlusAgentPlatform(agent_platform.upper()) if agent_platform else None
    resolved_git_provider = GitProvider(git_provider) if git_provider else None

    config = _resolve_config_with_prompts(
        agent_type=DgPlusAgentType.HYBRID,
        agent_platform=resolved_platform,
        organization=organization,
        url=url,
        deployment=deployment,
        git_root=git_root,
        python_version=python_version,
        skip_confirmation_prompt=skip_confirmation_prompt,
        use_editable_dagster=bool(use_editable_dagster),
        git_provider=resolved_git_provider,
        dg_context=dg_context,
        cli_config=cli_config,
        registry_url=registry_url,
    )

    configure_build_artifacts_impl(config)
    configure_ci_impl(config)


deploy_configure_group.add_command(deploy_configure_serverless)
deploy_configure_group.add_command(deploy_configure_hybrid)



"""Test run schema models in isolation.

These tests verify the basic data models work correctly without
importing any CLI or GraphQL components that might have import issues.
"""

import json

from dagster_dg_cli.api_layer.schemas.run import DgApiRun, DgApiRunStatus


class TestRunSchema:
    """Test the Run schema model."""

    def test_run_creation_minimal(self):
        """Test creating run with minimal required fields."""
        run = DgApiRun(
            id="test-run-123",
            status=DgApiRunStatus.QUEUED,
            created_at=1705311000.0,  # 2024-01-15T10:30:00Z
        )

        assert run.id == "test-run-123"
        assert run.status == DgApiRunStatus.QUEUED
        assert run.created_at == 1705311000.0
        assert run.started_at is None
        assert run.ended_at is None
        assert run.job_name is None

    def test_run_creation_complete(self):
        """Test creating run with all fields."""
        run = DgApiRun(
            id="complete-run-456",
            status=DgApiRunStatus.SUCCESS,
            created_at=1705311000.0,  # 2024-01-15T10:30:00Z
            started_at=1705311060.0,  # 2024-01-15T10:31:00Z
            ended_at=1705311900.0,  # 2024-01-15T10:45:00Z
            job_name="my_pipeline",
        )

        assert run.id == "complete-run-456"
        assert run.status == DgApiRunStatus.SUCCESS
        assert run.created_at == 1705311000.0
        assert run.started_at == 1705311060.0
        assert run.ended_at == 1705311900.0
        assert run.job_name == "my_pipeline"

    def test_run_json_serialization(self):
        """Test that Run can be serialized to JSON."""
        run = DgApiRun(
            id="json-test-789",
            status=DgApiRunStatus.FAILURE,
            created_at=1705311000.0,  # 2024-01-15T10:30:00Z
            started_at=1705311060.0,  # 2024-01-15T10:31:00Z
            ended_at=1705311180.0,  # 2024-01-15T10:33:00Z
            job_name="failing_pipeline",
        )

        json_str = run.model_dump_json()
        parsed = json.loads(json_str)

        assert parsed["id"] == "json-test-789"
        assert parsed["status"] == "FAILURE"
        assert parsed["created_at"] == 1705311000.0
        assert parsed["started_at"] == 1705311060.0
        assert parsed["ended_at"] == 1705311180.0
        assert parsed["job_name"] == "failing_pipeline"

    def test_run_json_deserialization(self):
        """Test that Run can be created from JSON."""
        json_data = {
            "id": "from-json-abc",
            "status": "STARTED",
            "created_at": 1705311000.0,  # 2024-01-15T10:30:00Z
            "started_at": 1705311060.0,  # 2024-01-15T10:31:00Z
            "ended_at": None,
            "job_name": "json_pipeline",
        }

        run = DgApiRun(**json_data)

        assert run.id == "from-json-abc"
        assert run.status == DgApiRunStatus.STARTED
        assert run.created_at == 1705311000.0
        assert run.started_at == 1705311060.0
        assert run.ended_at is None
        assert run.job_name == "json_pipeline"


class TestRunStatusEnum:
    """Test the RunStatus enum."""

    def test_run_status_values(self):
        """Test that all expected RunStatus values are available."""
        expected_statuses = [
            "QUEUED",
            "STARTING",
            "STARTED",
            "SUCCESS",
            "FAILURE",
            "CANCELING",
            "CANCELED",
        ]

        actual_statuses = [status.value for status in DgApiRunStatus]
        assert set(actual_statuses) == set(expected_statuses)

    def test_run_status_enum_creation(self):
        """Test creating runs with each status value."""
        for status in DgApiRunStatus:
            run = DgApiRun(
                id=f"test-{status.value.lower()}",
                status=status,
                created_at=1705311000.0,  # 2024-01-15T10:30:00Z
            )
            assert run.status == status

    def test_run_status_string_values(self):
        """Test specific string values of statuses."""
        assert DgApiRunStatus.QUEUED.value == "QUEUED"
        assert DgApiRunStatus.STARTING.value == "STARTING"
        assert DgApiRunStatus.STARTED.value == "STARTED"
        assert DgApiRunStatus.SUCCESS.value == "SUCCESS"
        assert DgApiRunStatus.FAILURE.value == "FAILURE"
        assert DgApiRunStatus.CANCELING.value == "CANCELING"
        assert DgApiRunStatus.CANCELED.value == "CANCELED"



"""Enhanced YAML template generator for component definitions.

This module provides functionality to generate comprehensive YAML templates
with schema documentation and example values for Dagster components.
"""

from typing import Any


def _get_field_description(field_schema: dict[str, Any]) -> str:
    """Extract field description from schema."""
    return field_schema.get("description", "")


def _resolve_ref(ref_path: str, full_schema: dict[str, Any]) -> dict[str, Any]:
    """Resolve a $ref path to its actual schema definition."""
    if not ref_path.startswith("#/$defs/"):
        return {}

    ref_name = ref_path.replace("#/$defs/", "")
    return full_schema.get("$defs", {}).get(ref_name, {})


def _get_example_value(
    field_schema: dict[str, Any], full_schema: dict[str, Any] | None = None
) -> Any:
    """Generate example values based on schema."""
    full_schema = full_schema or {}

    # Handle $ref objects
    if "$ref" in field_schema:
        ref_schema = _resolve_ref(field_schema["$ref"], full_schema)
        return _get_example_value(ref_schema, full_schema)

    # Handle anyOf schemas by preferring non-string types over string types
    # Do this BEFORE using schema examples to ensure we show the true underlying data structure
    if "anyOf" in field_schema:
        ref_options = []
        non_string_options = []
        string_options = []

        for option in field_schema["anyOf"]:
            if option.get("type") == "null":
                continue
            elif "$ref" in option:
                ref_options.append(option)
            elif option.get("type") == "string":
                string_options.append(option)
            else:
                non_string_options.append(option)

        # For union types with multiple $ref objects, use the first complex $ref option
        if len(ref_options) > 1:
            # Find the first $ref that resolves to a complex object with properties
            for ref_option in ref_options:
                ref_schema = _resolve_ref(ref_option["$ref"], full_schema)
                if ref_schema and ref_schema.get("properties"):
                    return _get_example_value(ref_schema, full_schema)
            # Fall back to first $ref if none have properties
            return _get_example_value(ref_options[0], full_schema)

        # Prefer $ref options, then non-string types, then string types
        all_options = ref_options + non_string_options + string_options
        if all_options:
            return _get_example_value(all_options[0], full_schema)

        # If all options are null, return null
        return None

    # Use schema-defined examples if available (for non-anyOf fields)
    if field_schema.get("examples"):
        return field_schema["examples"][0]

    field_type = field_schema.get("type")

    # Generic type-based examples
    if field_type == "string":
        return "example_string"
    elif field_type == "boolean":
        return True
    elif field_type == "integer":
        return 1
    elif field_type == "number":
        return 1.0
    elif field_type == "array":
        item_schema = field_schema.get("items", {})
        return [_get_example_value(item_schema, full_schema)]
    elif field_type == "object":
        properties = field_schema.get("properties", {})
        if properties:
            return {
                prop_name: _get_example_value(prop_schema, full_schema)
                for prop_name, prop_schema in properties.items()
            }
        elif field_schema.get("additionalProperties"):
            # Handle objects with additionalProperties but no specific properties
            additional_props_schema = field_schema.get("additionalProperties")
            if (
                isinstance(additional_props_schema, dict)
                and additional_props_schema.get("type") == "string"
            ):
                # Return a simple example object with string values
                return {"key": "value"}
            elif additional_props_schema is True:
                # additionalProperties: true means any properties are allowed
                return {"key": "value"}
        return {}

    return f"<{field_type or 'unknown'}>"


def _generate_schema_section(component_type: str, schema: dict[str, Any]) -> str:
    """Generate the schema documentation section as valid YAML with comments."""
    lines = []
    lines.append("# Template with instructions")
    lines.append(f"type: {component_type}  # Required")

    if "properties" in schema:
        lines.append("attributes:  # Optional: Attributes details")

        properties = schema["properties"]
        required_fields = set(schema.get("required", []))

        for field_name, field_schema in properties.items():
            _add_field_lines(lines, field_name, field_schema, required_fields, schema, indent="  ")
    else:
        lines.append("attributes:  # Optional: Attributes details")

    return "\n".join(lines)


def _generate_example_section(component_type: str, schema: dict[str, Any]) -> str:
    """Generate the example values section as valid YAML."""
    lines = []
    lines.append(f'type: "{component_type}"')

    if schema.get("properties"):
        lines.append("attributes:")

        properties = schema["properties"]
        for field_name, field_schema in properties.items():
            example_value = _get_example_value(field_schema, schema)
            _add_example_field_lines(lines, field_name, example_value, indent="  ")
    else:
        lines.append("attributes: {}")

    return "\n".join(lines)


def _generate_union_type_schema(
    lines: list[str],
    field_name: str,
    ref_options: list[dict[str, Any]],
    required_text: str,
    description: str,
    full_schema: dict[str, Any],
    indent: str = "",
) -> None:
    """Generate schema documentation for union types with multiple $ref objects."""
    lines.append(f"{indent}{field_name}:  # {required_text}: {description}")
    lines.append(f"{indent}  # Choose one of the following types:")

    for i, option in enumerate(ref_options, 1):
        ref_path = option["$ref"]
        ref_name = ref_path.split("/")[-1]
        ref_schema = _resolve_ref(ref_path, full_schema)

        if ref_schema and ref_schema.get("properties"):
            lines.append(f"{indent}  # Option {i} - {ref_name}:")
            sub_required = set(ref_schema.get("required", []))
            for sub_field_name, sub_field_schema in ref_schema["properties"].items():
                _add_field_lines(
                    lines,
                    f"# {sub_field_name}",
                    sub_field_schema,
                    sub_required,
                    full_schema,
                    indent + "  ",
                )
            lines.append(f"{indent}  #")


def _get_field_type_description(field_schema: dict[str, Any]) -> str:
    """Get a user-friendly type description for a field schema."""
    # Handle anyOf schemas
    if "anyOf" in field_schema:
        non_null_types = []
        for option in field_schema["anyOf"]:
            if option.get("type") != "null":
                if option.get("type") == "array":
                    item_type = option.get("items", {}).get("type", "item")
                    non_null_types.append(f"array of {item_type}")
                elif "$ref" in option:
                    ref_name = option["$ref"].split("/")[-1]
                    non_null_types.append(f"object ({ref_name})")
                else:
                    non_null_types.append(option.get("type", "unknown"))

        if non_null_types:
            if len(non_null_types) == 1:
                return non_null_types[0]
            else:
                return f"one of: {', '.join(non_null_types)}"

    # Handle regular type
    field_type = field_schema.get("type", "unknown")
    if field_type == "array":
        item_type = field_schema.get("items", {}).get("type", "item")
        return f"array of {item_type}"

    return field_type


def _add_field_lines(
    lines: list[str],
    field_name: str,
    field_schema: dict[str, Any],
    required_fields: set[str],
    full_schema: dict[str, Any],
    indent: str = "",
) -> None:
    """Add schema field lines with appropriate formatting and comments."""
    is_required = field_name in required_fields
    required_text = "Required" if is_required else "Optional"
    description = _get_field_description(field_schema)

    # Handle $ref objects - resolve and recurse
    if "$ref" in field_schema:
        ref_schema = _resolve_ref(field_schema["$ref"], full_schema)
        if ref_schema:
            _add_field_lines(lines, field_name, ref_schema, required_fields, full_schema, indent)
            return

    # Handle anyOf schemas - check for union types with multiple $ref objects
    if "anyOf" in field_schema:
        # Collect all $ref options (potential union type)
        ref_options = [option for option in field_schema["anyOf"] if "$ref" in option]

        # If we have multiple $ref options, treat as union type
        if len(ref_options) > 1:
            _generate_union_type_schema(
                lines, field_name, ref_options, required_text, description, full_schema, indent
            )
            return

        # Check if any anyOf option is a complex object that should be expanded (existing logic)
        complex_option = None
        for option in field_schema["anyOf"]:
            if option.get("type") == "array":
                items = option.get("items", {})
                if "$ref" in items:
                    # This is an array of complex objects - expand it
                    ref_schema = _resolve_ref(items["$ref"], full_schema)
                    if ref_schema and ref_schema.get("properties"):
                        lines.append(f"{indent}{field_name}:  # {required_text}: {description}")
                        lines.append(
                            f"{indent}  - # Array of {items['$ref'].split('/')[-1]} objects:"
                        )
                        sub_required = set(ref_schema.get("required", []))
                        for sub_field_name, sub_field_schema in ref_schema["properties"].items():
                            _add_field_lines(
                                lines,
                                sub_field_name,
                                sub_field_schema,
                                sub_required,
                                full_schema,
                                indent + "    ",
                            )
                        return
            elif "$ref" in option:
                ref_schema = _resolve_ref(option["$ref"], full_schema)
                if ref_schema and ref_schema.get("properties"):
                    complex_option = ref_schema
                    break

        if complex_option:
            # Expand the complex object option
            lines.append(f"{indent}{field_name}:  # {required_text}: {description}")
            sub_required = set(complex_option.get("required", []))
            for sub_field_name, sub_field_schema in complex_option["properties"].items():
                _add_field_lines(
                    lines,
                    sub_field_name,
                    sub_field_schema,
                    sub_required,
                    full_schema,
                    indent + "  ",
                )
        else:
            # Fall back to type description for simple anyOf
            type_desc = _get_field_type_description(field_schema)
            lines.append(f"{indent}{field_name}: <{type_desc}>  # {required_text}: {description}")
        return

    field_type = field_schema.get("type", "unknown")

    if field_type == "object":
        properties = field_schema.get("properties", {})
        if properties:
            lines.append(f"{indent}{field_name}:  # {required_text}: {description}")
            sub_required = set(field_schema.get("required", []))
            for sub_field_name, sub_field_schema in properties.items():
                _add_field_lines(
                    lines,
                    sub_field_name,
                    sub_field_schema,
                    sub_required,
                    full_schema,
                    indent + "  ",
                )
        else:
            lines.append(f"{indent}{field_name}:  # {required_text}: {description}")

    elif field_type == "array":
        item_schema = field_schema.get("items", {})

        # Handle array of complex objects via $ref
        if "$ref" in item_schema:
            ref_schema = _resolve_ref(item_schema["$ref"], full_schema)
            if ref_schema and ref_schema.get("properties"):
                lines.append(f"{indent}{field_name}:  # {required_text}: {description}")
                lines.append(
                    f"{indent}  - # Array of {item_schema['$ref'].split('/')[-1]} objects:"
                )
                sub_required = set(ref_schema.get("required", []))
                for sub_field_name, sub_field_schema in ref_schema["properties"].items():
                    _add_field_lines(
                        lines,
                        sub_field_name,
                        sub_field_schema,
                        sub_required,
                        full_schema,
                        indent + "    ",
                    )
                return

        # Handle regular arrays
        lines.append(
            f"{indent}{field_name}:  # {required_text}: List of {item_schema.get('type', 'item')} items"
        )
        if item_schema.get("type") == "object" and item_schema.get("properties"):
            lines.append(f"{indent}  - # List item properties:")
            item_required = set(item_schema.get("required", []))
            for item_field_name, item_field_schema in item_schema["properties"].items():
                _add_field_lines(
                    lines,
                    item_field_name,
                    item_field_schema,
                    item_required,
                    full_schema,
                    indent + "    ",
                )
        else:
            lines.append(f"{indent}  - <{item_schema.get('type', 'string')}>")

    else:
        lines.append(f"{indent}{field_name}: <{field_type}>  # {required_text}: {description}")


def _add_example_field_lines(
    lines: list[str], field_name: str, value: Any, indent: str = ""
) -> None:
    """Add example field lines with proper YAML formatting."""
    if isinstance(value, dict):
        lines.append(f"{indent}{field_name}:")
        for sub_key, sub_value in value.items():
            _add_example_field_lines(lines, sub_key, sub_value, indent + "  ")
    elif isinstance(value, list):
        lines.append(f"{indent}{field_name}:")
        for item in value:
            if isinstance(item, dict):
                lines.append(f"{indent}  -")
                for sub_key, sub_value in item.items():
                    _add_example_field_lines(lines, sub_key, sub_value, indent + "    ")
            else:
                lines.append(f'{indent}  - "{item}"')
    elif isinstance(value, str):
        lines.append(f'{indent}{field_name}: "{value}"')
    elif isinstance(value, bool):
        lines.append(f"{indent}{field_name}: {str(value).lower()}")
    else:
        lines.append(f"{indent}{field_name}: {value}")


def generate_defs_yaml_schema(component_type: str, schema: dict[str, Any]) -> str:
    """Generate YAML schema template with type information and documentation.

    Args:
        component_type: The component type identifier
        schema: The JSON schema for the component

    Returns:
        A YAML template string with schema documentation and type hints
    """
    if not schema:
        # Fallback for components without schema
        return f"""# Template with instructions
type: {component_type}  # Required
attributes:  # Optional: Attributes details"""

    return _generate_schema_section(component_type, schema)


def generate_defs_yaml_example_values(component_type: str, schema: dict[str, Any]) -> str:
    """Generate YAML example values with sample data.

    Args:
        component_type: The component type identifier
        schema: The JSON schema for the component

    Returns:
        A YAML string with example values and sample data
    """
    if not schema:
        # Fallback for components without schema
        return f"""type: "{component_type}"
attributes: {{}}"""

    return _generate_example_section(component_type, schema)



import webbrowser

import click
from dagster_dg_core.utils import DgClickCommand
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper
from dagster_shared.plus.config import DagsterPlusCliConfig

from dagster_dg_cli.utils.plus.gql import FULL_DEPLOYMENTS_QUERY
from dagster_dg_cli.utils.plus.gql_client import DagsterPlusGraphQLClient

EU_DAGSTER_CLOUD_URL = "https://eu.dagster.cloud"


@click.command(name="login", cls=DgClickCommand)
@click.option(
    "--region",
    type=click.Choice(["us", "eu"]),
    default=None,
    help="Dagster Cloud region to login to. Use 'eu' for European region.",
)
@cli_telemetry_wrapper
def login_command(region: str | None) -> None:
    """Login to Dagster Plus."""
    # Import login server only when the command is actually used
    from dagster_shared.plus.login_server import start_login_server

    # Determine the base URL based on region
    if region == "eu":
        org_url = EU_DAGSTER_CLOUD_URL
    elif DagsterPlusCliConfig.exists():
        org_url = DagsterPlusCliConfig.get().url
    else:
        org_url = None

    server, url = start_login_server(org_url)

    try:
        webbrowser.open(url, new=0, autoraise=True)
        click.echo(
            f"Opening browser...\nIf a window does not open automatically, visit {url} to"
            " finish authorization"
        )
    except webbrowser.Error as e:
        click.echo(f"Error launching web browser: {e}\n\nTo finish authorization, visit {url}\n")

    server.serve_forever()
    new_org = server.get_organization()
    new_api_token = server.get_token()

    config = DagsterPlusCliConfig(
        organization=new_org,
        user_token=new_api_token,
        url=org_url,
    )
    config.write()
    click.echo(f"Authorized for organization {new_org}\n")

    gql_client = DagsterPlusGraphQLClient.from_config(config)
    result = gql_client.execute(FULL_DEPLOYMENTS_QUERY)
    deployment_names = [d["deploymentName"] for d in result["fullDeployments"]]

    click.echo("Available deployments: " + ", ".join(deployment_names))

    selected_deployment = None
    while selected_deployment not in deployment_names:
        if selected_deployment is not None:
            click.echo(f"{selected_deployment} is not a valid deployment")
        selected_deployment = click.prompt(
            "Select a default deployment", default=deployment_names[0]
        )

    config = DagsterPlusCliConfig(
        organization=config.organization,
        user_token=config.user_token,
        default_deployment=selected_deployment,
        url=org_url,
    )
    config.write()



import sys
from contextlib import asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest
from dagster_shared import seven
from dagster_test.dg_utils.utils import ProxyRunner, isolated_example_project_foo_bar


@asynccontextmanager
async def mcp_server():
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.stdio import stdio_client

    server_params = StdioServerParameters(command="dg", args=["mcp", "serve"], env=None)

    async with stdio_client(server_params) as stdio_transport:
        stdio, write = stdio_transport
        async with ClientSession(stdio, write) as session:
            await session.initialize()
            yield session


@pytest.mark.skipif(sys.version_info < (3, 10), reason="no mcp support on 3.9")
@pytest.mark.asyncio
async def test_is_valid_mcp_server():
    async with mcp_server() as session:
        response = await session.list_tools()
        tools = response.tools
        assert "scaffold_dagster_component" in [tool.name for tool in tools]


# TODO: I would like to write a testing abstraction that lets us consolidate the mcp tests with the CLI tests,
# since the inputs are nearly identical in each case. For now, we just have basic coverage w/ duplicate tests.
@pytest.mark.skipif(sys.version_info < (3, 10), reason="no mcp support on 3.9")
@pytest.mark.skipif(
    seven.IS_PYTHON_3_14, reason="uses dagster_dbt, but dbt-core doesn't support 3.14"
)
@pytest.mark.asyncio
async def test_list_dagster_components():
    if TYPE_CHECKING:
        from mcp.types import TextContent
    with ProxyRunner.test() as runner, isolated_example_project_foo_bar(runner, in_workspace=False):
        async with mcp_server() as session:
            response = await session.call_tool("list_available_components", {"project_path": "."})
            assert not response.isError
            assert len(response.content) == 1
            text_content = cast("TextContent", response.content[0])
            assert "dagster.DefinitionsComponent" in text_content.text


@pytest.mark.skipif(sys.version_info < (3, 10), reason="no mcp support on 3.9")
@pytest.mark.skipif(
    seven.IS_PYTHON_3_14, reason="uses dagster_dbt, but dbt-core doesn't support 3.14"
)
@pytest.mark.asyncio
async def test_scaffold_dagster_component_and_check_yaml():
    with ProxyRunner.test() as runner, isolated_example_project_foo_bar(runner, in_workspace=False):
        async with mcp_server() as session:
            response = await session.call_tool(
                "scaffold_dagster_component",
                {
                    "project_path": ".",
                    "component_type": "dagster.DefinitionsComponent",
                    "component_name": "my_defs",
                    "component_arguments": [],
                },
            )
            assert not response.isError

            assert (Path.cwd() / "src" / "foo_bar/" / "defs" / "my_defs" / "defs.yaml").exists()

            response = await session.call_tool(
                "check_dagster_defs_yaml",
                {
                    "project_path": ".",
                },
            )
            assert response.isError

            assert (Path.cwd() / "src" / "foo_bar/" / "defs" / "my_defs" / "defs.yaml").write_text(
                "type: dagster.DefinitionsComponent\n\nattributes:\n  path: test.py"
            )

            response = await session.call_tool(
                "check_dagster_defs_yaml",
                {
                    "project_path": ".",
                },
            )
            assert not response.isError, response.content


@pytest.mark.skipif(sys.version_info < (3, 10), reason="no mcp support on 3.9")
@pytest.mark.asyncio
async def test_list_dagster_integrations():
    with ProxyRunner.test() as runner, isolated_example_project_foo_bar(runner, in_workspace=False):
        async with mcp_server() as session:
            response = await session.call_tool("list_available_integrations", {"project_path": "."})
            assert not response.isError
            assert len(response.content) == 1
            assert isinstance(response.content, list)
            assert dict(
                response.content[0]
            )  # check mcp.types.TextContent conforms to dict structure



"""Shared test utilities for API testing."""

from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest
from dagster_shared.plus.config import DagsterPlusCliConfig
from syrupy.extensions.amber import AmberSnapshotExtension

if TYPE_CHECKING:
    from syrupy.location import PyTestLocation
    from syrupy.types import SnapshotIndex


class ApiSnapshotExtension(AmberSnapshotExtension):
    """Custom snapshot extension for API tests.

    This extension ensures consistent snapshot naming by:
    1. Removing any existing parameterized test suffixes to avoid duplicate snapshots
    2. Applying a standardized index format that ensures all snapshots follow the pattern test_name[index]

    Without this extension, Syrupy's default naming might create inconsistent snapshot names
    or fail to handle indexing properly for these API tests.
    """

    @classmethod
    def get_snapshot_name(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        *,
        index: "SnapshotIndex",
        test_location: "PyTestLocation",
    ) -> str:
        snapshot_name = test_location.snapshot_name

        # Remove parameterized test suffixes to avoid duplicate snapshots
        variant_start_index = snapshot_name.find("[")
        if variant_start_index != -1:
            snapshot_name = snapshot_name[:variant_start_index]

        return f"{snapshot_name}[{index}]"


@pytest.fixture
def mock_dagster_plus_config():
    """Mock DagsterPlusCliConfig for testing."""
    config = Mock(spec=DagsterPlusCliConfig)
    config.org_name = "test-org"
    config.api_token = "test-token"
    config.endpoint = "https://test.dagster.cloud/graphql"
    return config


@pytest.fixture(autouse=True)
def disable_telemetry(monkeypatch, tmp_path):
    """Disable Dagster telemetry for API tests to prevent output interference."""
    # Set up a temporary DAGSTER_HOME with telemetry disabled
    dagster_home = tmp_path / "dagster_home"
    dagster_home.mkdir()

    # Create dagster.yaml with telemetry disabled
    dagster_yaml = dagster_home / "dagster.yaml"
    dagster_yaml.write_text("telemetry:\n  enabled: false\n")

    # Set the DAGSTER_HOME environment variable
    monkeypatch.setenv("DAGSTER_HOME", str(dagster_home))


@pytest.fixture
def snapshot(snapshot):
    """Configure snapshot fixture for API tests."""
    return snapshot.use_extension(ApiSnapshotExtension)



from pathlib import Path

import pytest
from dagster_dg_core.utils import activate_venv, discover_repo_root, is_windows, pushd
from dagster_shared.utils import environ
from dagster_test.components.test_utils.test_cases import BASIC_INVALID_VALUE, BASIC_MISSING_VALUE
from dagster_test.dg_utils.utils import (
    ProxyRunner,
    assert_runner_result,
    create_project_from_components,
    install_editable_dg_dev_packages_to_venv,
    isolated_example_project_foo_bar,
    isolated_example_workspace,
)


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_check_defs_workspace_context_success():
    dagster_git_repo_dir = str(discover_repo_root(Path(__file__)))
    with (
        ProxyRunner.test() as runner,
        isolated_example_workspace(runner, create_venv=True),
        environ({"DAGSTER_GIT_REPO_DIR": dagster_git_repo_dir}),
    ):
        result = runner.invoke_create_dagster(
            "project",
            "--use-editable-dagster",
            "projects/project-1",
            "--uv-sync",
        )
        assert_runner_result(result)
        result = runner.invoke_create_dagster(
            "project",
            "--use-editable-dagster",
            "projects/project-2",
            "--uv-sync",
        )
        assert_runner_result(result)

        result = runner.invoke("check", "defs")
        assert_runner_result(result)

        (Path("projects") / "project-1" / "src" / "project_1" / "defs" / "__init__.py").write_text(
            "invalid"
        )
        result = runner.invoke("check", "defs")
        assert result.exit_code == 1


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_check_defs_project_context_success():
    with ProxyRunner.test() as runner, isolated_example_project_foo_bar(runner):
        result = runner.invoke("check", "defs")
        assert_runner_result(result)


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_check_defs_project_context_failure():
    with ProxyRunner.test() as runner, isolated_example_project_foo_bar(runner):
        (Path("src") / "foo_bar" / "defs" / "__init__.py").write_text("invalid")
        result = runner.invoke("check", "defs")
        assert result.exit_code == 1


def test_implicit_yaml_check_from_dg_check_defs_in_project_context() -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_MISSING_VALUE.component_path,
            BASIC_INVALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_MISSING_VALUE.component_type_filepath,
            in_workspace=False,
        ) as tmpdir,
    ):
        with pushd(str(tmpdir)):
            result = runner.invoke("check", "defs", catch_exceptions=False)
            assert result.exit_code != 0, str(result.output)

            assert BASIC_INVALID_VALUE.check_error_msg and BASIC_MISSING_VALUE.check_error_msg
            BASIC_INVALID_VALUE.check_error_msg(str(result.output))
            BASIC_MISSING_VALUE.check_error_msg(str(result.output))

            # didn't make it to defs check
            assert "Validation failed for code location foo-bar" not in str(result.output)


def test_implicit_yaml_check_disabled_in_project_context() -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_MISSING_VALUE.component_path,
            BASIC_INVALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_MISSING_VALUE.component_type_filepath,
            in_workspace=False,
        ) as tmpdir,
    ):
        with pushd(str(tmpdir)):
            result = runner.invoke("check", "defs", "--no-check-yaml", catch_exceptions=False)
            assert result.exit_code != 0, str(result.output)

            # yaml check was skipped, defs check failed
            assert "Validation failed for code location foo-bar" in str(result.output)


def test_no_implicit_yaml_check_from_dg_check_defs_in_workspace_context() -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_MISSING_VALUE.component_path,
            BASIC_INVALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_MISSING_VALUE.component_type_filepath,
            in_workspace=True,
        ) as tmpdir,
    ):
        with pushd(Path(tmpdir).parent):
            result = runner.invoke("check", "defs", catch_exceptions=False)
            assert result.exit_code != 0, str(result.output)

            # yaml check was skipped, defs check failed
            assert "Validation failed for code location foo-bar" in str(result.output)


def test_implicit_yaml_check_from_dg_check_defs_disallowed_in_workspace_context() -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_MISSING_VALUE.component_path,
            BASIC_INVALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_MISSING_VALUE.component_type_filepath,
            in_workspace=True,
        ) as tmpdir,
    ):
        with pushd(Path(tmpdir).parent):
            result = runner.invoke("check", "defs", "--check-yaml", catch_exceptions=False)
            assert result.exit_code != 0, str(result.output)

            assert "--check-yaml is not currently supported in a workspace context" in str(
                result.output
            )

        # It is supported and is the default in a project context within a workspace
        with pushd(tmpdir):
            result = runner.invoke("check", "defs", "--check-yaml", catch_exceptions=False)
            assert result.exit_code != 0, str(result.output)

            assert BASIC_INVALID_VALUE.check_error_msg and BASIC_MISSING_VALUE.check_error_msg
            BASIC_INVALID_VALUE.check_error_msg(str(result.output))
            BASIC_MISSING_VALUE.check_error_msg(str(result.output))


@pytest.mark.skipif(is_windows(), reason="Temporarily skipping (signal issues in CLI)..")
def test_check_defs_uses_active_venv_when_flag_set():
    """Test that check defs command logs the active venv Python when --use-active-venv is set."""
    dagster_git_repo_dir = str(discover_repo_root(Path(__file__)))
    with (
        ProxyRunner.test() as runner,
        isolated_example_workspace(runner, create_venv=True) as workspace_path,
        environ({"DAGSTER_GIT_REPO_DIR": dagster_git_repo_dir}),
    ):
        venv_path = workspace_path / ".venv"
        install_editable_dg_dev_packages_to_venv(venv_path)

        with activate_venv(venv_path):
            # Create a project
            result = runner.invoke_create_dagster(
                "project", "--use-editable-dagster", "test-project", "--uv-sync"
            )
            assert_runner_result(result)

            # Run check defs with --use-active-venv flag
            result = runner.invoke("check", "defs", "--use-active-venv")

            # Verify the command succeeded
            assert_runner_result(result)

            # Verify the log message is present in output
            assert "Using active Python environment:" in str(result.output), (
                f"Expected log message about using active Python environment, but got:\n{result.output}"
            )



"""Client factory for DG API commands."""

from typing import Protocol

import click
from dagster_shared.plus.config import DagsterPlusCliConfig

from dagster_dg_cli.utils.plus.gql_client import DagsterPlusGraphQLClient, IGraphQLClient

# Test constants
TEST_ORGANIZATION = "test-org"
TEST_DEPLOYMENT = "test-deployment"


class GraphQLClientFactory(Protocol):
    """Protocol for GraphQL client factories used in testing."""

    def __call__(self, config: DagsterPlusCliConfig) -> IGraphQLClient: ...


class DgApiTestContext:
    """Test context for DG API commands."""

    def __init__(self, client_factory: GraphQLClientFactory):
        self.client_factory = client_factory
        self.organization = TEST_ORGANIZATION
        self.deployment = TEST_DEPLOYMENT


def create_dg_api_graphql_client(
    ctx: click.Context, config: DagsterPlusCliConfig, view_graphql: bool = False
) -> IGraphQLClient:
    """Create GraphQL client for DG API commands.

    This is the single entry point for GraphQL client creation in DG API commands.
    It checks for test context injection first, then handles authentication for normal usage.

    Args:
        ctx: Click context from the CLI command
        config: Config to use for creating the client
        view_graphql: If True, wrap the client with debugging output

    Returns:
        IGraphQLClient instance
    """
    # Check if we have a test context with custom factory
    if ctx.obj and isinstance(ctx.obj, DgApiTestContext) and ctx.obj.client_factory:
        client = ctx.obj.client_factory(config)
    else:
        # For normal operation, validate token exists and create client
        if not config.user_token:
            raise click.UsageError(
                "A Dagster Cloud API token must be specified.\n\n"
                "You may specify a token by:\n"
                "- Providing the --api-token parameter\n"
                "- Setting the DAGSTER_CLOUD_API_TOKEN environment variable"
            )

        # Normal operation: create real client from config
        client = DagsterPlusGraphQLClient.from_config(config)

    # Wrap with debug client if requested
    if view_graphql:
        from dagster_dg_cli.utils.plus.gql_client import DebugGraphQLClient

        client = DebugGraphQLClient(client)

    return client


def create_dg_api_client(ctx: click.Context) -> IGraphQLClient:
    """Create GraphQL client for DG API commands with automatic config handling.

    This is a convenience function for deployment commands that handles both
    config creation and client creation in a single step.

    Args:
        ctx: Click context from the CLI command

    Returns:
        IGraphQLClient instance
    """
    from dagster_dg_cli.cli.api.shared import get_config_for_api_command

    config = get_config_for_api_command(ctx)
    return create_dg_api_graphql_client(ctx, config)



"""Test asset business logic functions without mocks.

These tests focus on testing pure functions that process data without requiring
GraphQL client mocking or external dependencies.
"""

from dagster_dg_cli.api_layer.schemas.asset import DgApiAsset, DgApiAssetList
from dagster_dg_cli.cli.api.formatters import format_asset, format_assets


class TestFormatAssets:
    """Test the asset formatting functions."""

    def _create_sample_asset_list(self):
        """Create sample DgApiAssetList for testing."""
        assets = [
            DgApiAsset(
                id="asset-1",
                asset_key="users",
                asset_key_parts=["users"],
                description="User data table",
                group_name="core_tables",
                kinds=["dbt", "table"],
                metadata_entries=[
                    {"label": "owner", "description": "data-team"},
                    {"label": "rows", "description": "1000000", "intValue": 1000000},
                ],
            ),
            DgApiAsset(
                id="asset-2",
                asset_key="analytics/user_metrics",
                asset_key_parts=["analytics", "user_metrics"],
                description="Aggregated user metrics",
                group_name="analytics",
                kinds=["dbt", "view"],
                metadata_entries=[],
            ),
        ]
        return DgApiAssetList(items=assets, cursor="next_cursor", has_more=True)

    def _create_empty_asset_list(self):
        """Create empty DgApiAssetList for testing."""
        return DgApiAssetList(items=[], cursor=None, has_more=False)

    def _create_single_asset(self):
        """Create single DgApiAsset for testing."""
        return DgApiAsset(
            id="asset-1",
            asset_key="users",
            asset_key_parts=["users"],
            description="User data table",
            group_name="core_tables",
            kinds=["dbt", "table"],
            metadata_entries=[
                {"label": "owner", "description": "data-team"},
                {"label": "schema", "description": "public", "text": "public"},
            ],
        )

    def test_format_assets_text_output(self, snapshot):
        """Test formatting assets as text."""
        asset_list = self._create_sample_asset_list()
        result = format_assets(asset_list, as_json=False)

        # Snapshot the entire text output
        snapshot.assert_match(result)

    def test_format_assets_json_output(self, snapshot):
        """Test formatting assets as JSON."""
        asset_list = self._create_sample_asset_list()
        result = format_assets(asset_list, as_json=True)

        # For JSON, we want to snapshot the parsed structure to avoid formatting differences
        import json

        parsed = json.loads(result)
        snapshot.assert_match(parsed)

    def test_format_empty_assets_text_output(self, snapshot):
        """Test formatting empty asset list as text."""
        asset_list = self._create_empty_asset_list()
        result = format_assets(asset_list, as_json=False)

        snapshot.assert_match(result)

    def test_format_empty_assets_json_output(self, snapshot):
        """Test formatting empty asset list as JSON."""
        asset_list = self._create_empty_asset_list()
        result = format_assets(asset_list, as_json=True)

        import json

        parsed = json.loads(result)
        snapshot.assert_match(parsed)

    def test_format_single_asset_text_output(self, snapshot):
        """Test formatting single asset as text."""
        asset = self._create_single_asset()
        result = format_asset(asset, as_json=False)

        snapshot.assert_match(result)

    def test_format_single_asset_json_output(self, snapshot):
        """Test formatting single asset as JSON."""
        asset = self._create_single_asset()
        result = format_asset(asset, as_json=True)

        import json

        parsed = json.loads(result)
        snapshot.assert_match(parsed)

    def test_format_asset_without_metadata(self, snapshot):
        """Test formatting asset with no metadata."""
        asset = DgApiAsset(
            id="simple-asset",
            asset_key="simple",
            asset_key_parts=["simple"],
            description=None,
            group_name="default",
            kinds=[],
            metadata_entries=[],
        )
        result = format_asset(asset, as_json=False)

        snapshot.assert_match(result)


class TestAssetDataProcessing:
    """Test processing of asset data structures.

    This class would test any pure functions in the GraphQL adapter
    that process the raw GraphQL responses into our domain models.
    Since the actual GraphQL processing is done inline in the adapter
    functions, these tests will verify our data model creation.
    """

    def test_asset_creation_with_complex_metadata(self, snapshot):
        """Test creating asset with various metadata types."""
        asset = DgApiAsset(
            id="complex-asset",
            asset_key="analytics/complex_table",
            asset_key_parts=["analytics", "complex_table"],
            description="Complex asset with various metadata types",
            group_name="analytics",
            kinds=["dbt", "table", "materialized"],
            metadata_entries=[
                {"label": "text_meta", "description": "Some text", "text": "example_text"},
                {
                    "label": "url_meta",
                    "description": "Documentation",
                    "url": "https://docs.example.com",
                },
                {
                    "label": "path_meta",
                    "description": "File path",
                    "path": "/data/warehouse/table.sql",
                },
                {"label": "json_meta", "description": "Config", "jsonString": '{"key": "value"}'},
                {
                    "label": "markdown_meta",
                    "description": "Notes",
                    "mdStr": "# Notes\nThis is markdown",
                },
                {
                    "label": "python_meta",
                    "description": "Python class",
                    "module": "mymodule",
                    "name": "MyClass",
                },
                {"label": "float_meta", "description": "Score", "floatValue": 95.5},
                {"label": "int_meta", "description": "Count", "intValue": 42},
                {"label": "bool_meta", "description": "Is active", "boolValue": True},
            ],
        )

        # Test JSON serialization works correctly
        result = asset.model_dump_json(indent=2)
        import json

        parsed = json.loads(result)
        snapshot.assert_match(parsed)

    def test_nested_asset_key_parsing(self):
        """Test asset key parsing with nested paths."""
        asset = DgApiAsset(
            id="nested-asset",
            asset_key="level1/level2/level3/asset",
            asset_key_parts=["level1", "level2", "level3", "asset"],
            description="Deeply nested asset",
            group_name="nested_group",
            kinds=["table"],
            metadata_entries=[],
        )

        assert asset.asset_key == "level1/level2/level3/asset"
        assert asset.asset_key_parts == ["level1", "level2", "level3", "asset"]
        assert len(asset.asset_key_parts) == 4



"""Scaffold submodule for dagster-dg CLI."""

import click
from dagster_dg_core.utils import DgClickGroup

from dagster_dg_cli.cli.scaffold.branch.command import scaffold_branch_command
from dagster_dg_cli.cli.scaffold.build_artifacts import scaffold_build_artifacts_command
from dagster_dg_cli.cli.scaffold.component import scaffold_component_command
from dagster_dg_cli.cli.scaffold.defs import scaffold_defs_group
from dagster_dg_cli.cli.scaffold.github_actions import scaffold_github_actions_command


@click.group(name="scaffold", cls=DgClickGroup)
def scaffold_group():
    """Commands for scaffolding Dagster entities."""


# Register all scaffold commands from submodules
scaffold_group.add_command(scaffold_branch_command)
scaffold_group.add_command(scaffold_build_artifacts_command)  # Legacy shim
scaffold_group.add_command(scaffold_component_command)
scaffold_group.add_command(scaffold_defs_group)
scaffold_group.add_command(scaffold_github_actions_command)  # Legacy shim
